import logging
import pandas as pd
import numpy as np
from multiprocessing import Pool
from collections import defaultdict, Counter, namedtuple

from CIRI_long import env
from CIRI_long.align import *
from CIRI_long.utils import *
from CIRI_long.logger import ProgressBar


LOGGER = logging.getLogger('CIRI-long')
READ = namedtuple('Read', 'read_id circ_id strand cirexon ss clip segments seq sample type')
CIRC = namedtuple('Circ', 'contig start end strand')
globals()['Read'] = READ
globals()['Circ'] = CIRC


class Segment(object):
    def __init__(self, start, end):
        self.start = start
        self.end = end
        self.last = None
        self.next = None

    def __str__(self):
        return '{}-{}'.format(self.start, self.end)


class Exon(Segment):
    def __init__(self, start, end):
        self.start = int(start)
        self.end = int(end)


def load_cand_circ(in_file):
    from pathlib import Path
    sample_attr = {}
    with open(in_file, 'r') as f:
        for line in f:
            content = line.rstrip().split()
            if content:
                sample, fname = content
                sample_attr[sample] = fname

    cand_reads = {}
    for sample, fname in sample_attr.items():
        # Full length reads
        cand_circ = Path(fname)
        with open(cand_circ, 'r') as f:
            for line in f:
                content = line.rstrip().lstrip('>').split('\t')
                clip_base = int(content[5].split('|')[1].split('-')[0])
                seq = f.readline().rstrip()
                if clip_base > 20:
                    continue
                cand_reads[content[0]] = READ(*content, seq, sample, 'full')

        # Partial reads
        prefix = cand_circ.name.split('.')[0]
        with open(cand_circ.parent / (prefix + '.low_confidence.fa')) as f:
            for line in f:
                content = line.rstrip().lstrip('>').split('\t')
                clip_base = int(content[5].split('|')[1].split('-')[0])
                seq = f.readline().rstrip()
                if clip_base > 20:
                    continue
                cand_reads[content[0]] = READ(*content, seq, sample, 'partial')

    return cand_reads


def cluster_reads(cand_reads):
    import re
    from operator import itemgetter

    circ_reads = defaultdict(list)
    circ_start = defaultdict(dict)
    circ_end = defaultdict(dict)

    # Load circRNA junction sites
    for read_id, read in cand_reads.items():
        contig, start, end = re.split('[:-]', read.circ_id)
        start, end = int(start), int(end)
        if end - start > 200 * 1000:
            continue

        # Store reads & BSJ sites
        circ_reads[contig].append((start, end, read.read_id))
        circ_start[contig].setdefault(start, []).append(read.read_id)
        circ_end[contig].setdefault(end, []).append(read.read_id)

    # Build index of circRNA junction sites
    reads_cluster = []
    for contig in circ_reads:
        circ_start_index = {}
        circ_end_index = {}

        # Cluster of start junction site
        tmp = [[], ]
        for x in sorted(circ_start[contig]):
            if not tmp[-1]:
                tmp[-1].append(x)
            elif x > tmp[-1][-1] + 20:
                tmp.append([x, ])
            else:
                tmp[-1].append(x)
        for x in tmp:
            for i in range(min(x) // 500, max(x) // 500 + 1):
                circ_start_index.setdefault(i, []).append(x)

        # Cluster of end junction site
        tmp = [[], ]
        for x in sorted(circ_end[contig]):
            if not tmp[-1]:
                tmp[-1].append(x)
            elif x > tmp[-1][-1] + 20:
                tmp.append([x, ])
            else:
                tmp[-1].append(x)
        for x in tmp:
            for i in range(min(x) // 500, max(x) // 500 + 1):
                circ_end_index.setdefault(i, []).append(x)

        # Cluster reads
        reads_itered = {}
        for (start, end, read_id) in sorted(circ_reads[contig], key=itemgetter(0, 1)):
            if read_id in reads_itered:
                continue

            tmp_reads = []
            p = [i for i in circ_start_index[start // 500] if start in i][0]
            q = [i for i in circ_end_index[end // 500] if end in i][0]

            for i in p:
                tmp_start = circ_start[contig][i]
                for j in q:
                    tmp_end = circ_end[contig][j]
                    tmp = set(tmp_start) & set(tmp_end)
                    if tmp:
                        tmp_reads += tmp

            for i in tmp_reads:
                reads_itered[i] = 1

            reads_cluster.append(sorted([cand_reads[i] for i in tmp_reads], key=lambda x: len(x.seq), reverse=True))

    return reads_cluster


def genome_junction_seq(contig, start, end, width=25):
    return env.GENOME.seq(contig, end-width, end) + env.GENOME.seq(contig, start, start + width)


def avg_score(alignment, ref, query):
    from Levenshtein import distance
    x = query[alignment.query_begin:alignment.query_end]
    return distance(ref, x) / len(ref)


def curate_junction(ctg, st, en, junc):
    from libs.striped_smith_waterman.ssw_wrap import Aligner
    from operator import itemgetter
    junc_scores = []
    for i in range(max(0, min(st) - 25), max(st) + 25):
        for j in range(min(en) - 25, min(max(en) + 25, env.CONTIG_LEN[ctg])):
            if j <= i:
                continue
            tmp = genome_junction_seq(ctg, i, j, width=10)
            tmp_aligner = Aligner(tmp, match=10, mismatch=4, gap_open=8, gap_extend=2)
            tmp_score = avg_score(tmp_aligner.align(junc), tmp, junc)
            junc_scores.append((i, j, tmp_score))
    return sorted(junc_scores, key=itemgetter(2))


def annotated_hit(contig, scores):
    if contig not in env.SS_INDEX:
        return None

    weighted = []
    for st, en, score in scores:
        w = 0
        if st + 1 in env.SS_INDEX[contig]:
            tmp = set(flatten([p for _, p in env.SS_INDEX[contig][st + 1].items()]))
            if 'start' in tmp:
                w += 1
        elif st in env.SS_INDEX[contig]:
            tmp = set(flatten([p for _, p in env.SS_INDEX[contig][st].items()]))
            if 'end' in tmp:
                w += 1
        else:
            pass

        if en in env.SS_INDEX[contig]:
            tmp = set(flatten([p for _, p in env.SS_INDEX[contig][en].items()]))
            if 'end' in tmp:
                w += 1
        elif en + 1 in env.SS_INDEX[contig]:
            tmp = set(flatten([p for _, p in env.SS_INDEX[contig][en + 1].items()]))
            if 'start' in tmp:
                w += 1
        else:
            pass

        weighted.append([st, en, w])

    return min_sorted_items(weighted, 2, True)


def junc_score(ctg, junc, junc_seqs):
    from libs.striped_smith_waterman.ssw_wrap import Aligner
    aligner = Aligner(env.GENOME.seq(ctg, junc[0], junc[1]) * 2,
                      match=10, mismatch=4, gap_open=8, gap_extend=2)
    score = np.mean([aligner.align(i).score for i in junc_seqs])
    return score


def correct_chunk(chunk, max_cluster=200):
    cs_cluster = []
    cnt = defaultdict(int)
    for cluster in chunk:
        if cluster is None:
            continue
        # if '631ecb01-6f74-4de9-b8ab-c673b95cc4d3' not in [i.read_id for i in cluster]:
        #     continue
        ret = correct_cluster(cluster, max_cluster=max_cluster)
        if ret is None:
            continue
        circ_type, circ_attr = ret
        cnt[circ_type] += 1
        cs_cluster.append(circ_attr)
    return cs_cluster, cnt


def correct_cluster(cluster, is_debug=False, max_cluster=200):
    from random import sample
    from collections import Counter
    from spoa import poa
    from libs.striped_smith_waterman.ssw_wrap import Aligner

    if cluster is None:
        return None
    if len(cluster) <= 1:
        return None
    if 'full' not in set([i.type for i in cluster]):
        return None

    counter = Counter([i.circ_id for i in cluster if i.type == 'full']).most_common(n=1)
    ref = sorted([i for i in cluster if i.circ_id == counter[0][0] and i.type == 'full'],
                 key=lambda x: len(x.seq), reverse=True)[0]
    ssw = Aligner(ref.seq[:50], match=10, mismatch=4, gap_open=8, gap_extend=2)

    head_pos = []
    for query in cluster[1:]:
        alignment = ssw.align(query.seq)
        head_pos.append(alignment.ref_begin)

    template = transform_seq(ref.seq, max(head_pos))
    ssw = Aligner(template, match=10, mismatch=4, gap_open=8, gap_extend=2)
    junc_seqs = [get_junc_seq(template, -max(head_pos)//2, 25), ]

    for query in cluster[1:]:
        alignment = ssw.align(query.seq)
        tmp = transform_seq(query.seq, alignment.query_begin)
        junc_seqs.append(get_junc_seq(tmp, -max(head_pos)//2, 25))

    cs_junc, _ = poa(junc_seqs, 2, False, 10, -4, -8, -2, -24, -1)

    ctg = Counter([i.circ_id.split(':')[0] for i in cluster]).most_common()[0][0]
    tmp_st = [int(i.circ_id.split(':')[1].split('-')[0]) for i in cluster]
    tmp_en = [int(i.circ_id.split(':')[1].split('-')[1]) for i in cluster]

    # Curate junction sequence
    scores = curate_junction(ctg, tmp_st, tmp_en, cs_junc)
    aval_junc = min_sorted_items(scores, 2)
    if aval_junc:
        anno_junc = annotated_hit(ctg, aval_junc)
        if anno_junc:
            anno_junc = sorted(anno_junc, key=lambda x: junc_score(ctg, x, junc_seqs), reverse=True)
            circ_start, circ_end, circ_score = anno_junc[0]
        else:
            aval_junc = sorted(aval_junc, key=lambda x: junc_score(ctg, x, junc_seqs), reverse=True)
            circ_start, circ_end, circ_score = aval_junc[0]
    else:
        circ_start, circ_end = counter[0][0].split(':')[1].split('-')
        circ_start, circ_end = int(circ_start), int(circ_end)

    # Annotated sites
    for shift_threshold in [5, 10]:
        ss_site, us_free, ds_free, tmp_signal = find_annotated_signal(ctg, circ_start, circ_end, 0, 10, shift_threshold)
        if ss_site is not None:
            ss_id, strand, us_shift, ds_shift = ss_site
            circ_start += us_shift
            circ_end += ds_shift
            circ_type = 'Annotated'
            break

    host_strand = find_host_gene(ctg, circ_start, circ_end)
    circ_type = None

    # Canonical sites
    if ss_site is None:
        for shift_threshold in [5, 10]:
            ss_site = find_denovo_signal(ctg, circ_start, circ_end, host_strand, tmp_signal,
                                         us_free, ds_free, 0, 10, shift_threshold, True)
            if ss_site is not None:
                ss_id, strand, us_shift, ds_shift = ss_site
                circ_start += us_shift
                circ_end += ds_shift
                circ_type = 'Annotated'
                break

    # Intronic circRNAs
    if ss_site is None:
        retained_introns = find_retained_introns(ctg, circ_start + 1, circ_end)
        overlap_exons = find_overlap_exons(ctg, circ_start + 1, circ_end)

        is_lariat = 0
        if retained_introns is not None and overlap_exons is None:
            is_lariat = 1
            # Find high-confidence ciRNAs
            retained_introns = set(sum([i for _, i in retained_introns.items()], []))
            retained_strand = set([i[2] for i in retained_introns])
            tmp_circ = []
            for intron_start, intron_end, intron_strand in retained_introns:
                if abs(intron_start - circ_start) > 50 or abs(intron_end - circ_end) > 50:
                    continue
                if intron_strand == '+':
                    tmp_site = [i for i in scores if i[0] == intron_start]
                else:
                    tmp_site = [i for i in scores if i[1] == intron_end]
                if tmp_site:
                    tmp_circ.append([*tmp_site[0], intron_strand])

            ss_id = 'lariat'
            if tmp_circ:
                circ_start, circ_end, circ_score, strand = sorted(tmp_circ, key=lambda x: x[2])[0]
                circ_type = 'High confidence lariat'
            else:
                # Lariat with recursive splicing branchpoint
                is_lariat = 0
                tmp_circ = []
                for tmp_strand in retained_strand:
                    tmp_start, tmp_end, tmp_score = recursive_splice_site(scores, ctg, tmp_strand)
                    if tmp_score is not None:
                        tmp_circ.append([tmp_start, tmp_end, tmp_score, tmp_strand])
                if tmp_circ:
                    circ_start, circ_end, circ_score, strand = sorted(tmp_circ, key=lambda x: x[2])[0]
                    # cnt['Recursive splicing lariat'] += 1
                else:
                    # cnt['Unknown lariat'] += 1
                    strand = 'None'

        # Find denovo splice signal
        if is_lariat == 0:
            ss_site = find_denovo_signal(ctg, circ_start, circ_end, host_strand, tmp_signal,
                                         us_free, ds_free, 5, 10, 3, False)
            if ss_site is not None:
                ss_id, strand, us_shift, ds_shift = ss_site
                circ_start += us_shift
                circ_end += ds_shift
                circ_type = 'Denovo signal'
            else:
                ss_id = 'None'
                strand = 'None'
                circ_type = 'Unknown signal'

    circ_id = '{}:{}-{}'.format(ctg, circ_start + 1, circ_end)

    # refined sequence
    cluster_seq = []
    circ_junc_seq = genome_junction_seq(ctg, circ_start, circ_end)
    ssw = Aligner(circ_junc_seq, match=10, mismatch=4, gap_open=8, gap_extend=2, report_cigar=True)

    tmp_cluster = [i for i in cluster if i.type == 'full']
    if len(tmp_cluster) > max_cluster:
        tmp_cluster = sample(tmp_cluster, max_cluster)
    tmp_cluster = sorted(tmp_cluster, key=lambda x: len(x.seq), reverse=True)

    for query in tmp_cluster:
        alignment = ssw.align(query.seq * 2)
        tmp_pos = find_alignment_pos(alignment, len(circ_junc_seq)//2)
        if tmp_pos is None:
            cluster_seq.append((query.read_id, query.seq ))
        else:
            tmp_seq = transform_seq(query.seq, tmp_pos % len(query.seq ))
            cluster_seq.append((query.read_id, tmp_seq))

    cluster_res = batch_cluster_sequence(circ_id, cluster_seq)
    cluster_res = sorted(cluster_res, key=lambda x: len(x[1]), reverse=True)

    circ = CIRC(ctg, circ_start + 1, circ_end, strand)
    circ_id = '{}:{}-{}'.format(circ.contig, circ.start, circ.end)

    if len(cluster_res) > 2 and len(cluster_res[0][1]) >= 0.5 * max(len(tmp_cluster), 10):
        tmp_res = correct_cluster([i for i in cluster if i.read_id in cluster_res[0][1]], True)
        if tmp_res is not None:
            circ = tmp_res
            circ_id = '{}:{}-{}'.format(circ.contig, circ.start, circ.end)

    # Filter out strange cirexons
    curated_exons = curate_cirexons(circ, cluster)
    if curated_exons is None:
        return None
    isoforms, circ_len = curate_isoform(circ, curated_exons, cluster_res)
    if isoforms is None:
        return None
    is_concordance = check_isoforms(circ, isoforms)
    if not is_concordance:
        return None

    if is_debug:
        return circ

    return circ_type, ([i.read_id for i in cluster], cluster_seq, circ_id, circ.strand,
                       ss_id, us_free, ds_free, circ_len, isoforms)


def batch_cluster_sequence(circ_id, x):
    from CIRI_long.find_ccs import compress_seq
    sequence = {}
    hpc_freq = []
    for read_id, read_seq in x:
        sequence[read_id] = read_seq
        hpc_freq.append((compress_seq(read_seq), [read_id, ]))

    res = iter_cluster_sequence(circ_id, hpc_freq, sequence)

    # If consensus
    for _ in range(10):
        n_res = cluster_sequence(res, sequence)
        if len(n_res) == len(res):
            break
        res = n_res
    else:
        LOGGER.warn('Sequence not consensus for circRNA: {}'.format(circ_id))
    return res


def iter_cluster_sequence(circ_id, hpc_freq, sequence):
    if len(hpc_freq) <= 50:
        return cluster_sequence(hpc_freq, sequence)

    res = []
    for tmp in grouper(hpc_freq, 50):
        chunk = [i for i in tmp if i is not None]
        res = cluster_sequence(chunk + res, sequence)

        for _ in range(10):
            n_res = cluster_sequence(res, sequence)
            if len(n_res) == len(res):
                break
            res = n_res
        else:
            LOGGER.warn('Sequence not consensus for circRNA: {}'.format(circ_id))
    return res


def cluster_sequence(hpc_freq, sequence):
    from scipy.cluster.hierarchy import linkage, leaves_list
    from scipy.spatial.distance import squareform
    from Levenshtein import distance
    from spoa import poa

    if len(hpc_freq) == 1:
        return hpc_freq

    # Calculate distance between segments
    dist = np.zeros((len(hpc_freq), len(hpc_freq)))
    for i in range(len(hpc_freq)):
        for j in range(len(hpc_freq)):
            if i > j:
                continue
            dist[i][j] = distance(hpc_freq[i][0], hpc_freq[j][0]) / max(len(hpc_freq[i][0]), len(hpc_freq[j][0]))
    dist = dist + dist.T

    # Hierarchical Cluster
    if dist.sum() != 0:
        z = leaves_list(linkage(squareform(dist), "ward", optimal_ordering=True))
    else:
        z = list(range(len(hpc_freq)))

    clusters = [[z[0], ]]
    for i, j in pairwise(z):
        if i > j:
            if dist[j][i] < 0.3:
                clusters[-1].append(j)
            else:
                clusters.append([j, ])
        else:
            if dist[i][j] < 0.3:
                clusters[-1].append(j)
            else:
                clusters.append([j, ])

    ccs_seq = []
    for cluster in clusters:
        if len(cluster) == 1:
            ccs_seq.append((hpc_freq[cluster[0]]))
            continue

        cluster_reads = flatten([hpc_freq[i][1] for i in cluster])
        cluster_seq = [sequence[i] for i in cluster_reads]

        # Generate consensus sequence
        ccs, _ = poa(cluster_seq, 2, False, 10, -4, -8, -2, -24, -1)
        ccs_seq.append((ccs, cluster_reads))
    return ccs_seq


def find_intron_signal(contig, starts, ends):
    if env.SS_INDEX is not None and contig in env.SS_INDEX:
        return None

    for strand in ['+', '-']:
        tmp_us_sites = []
        for st in range(min(starts)-10, max(starts)+10):
            us_pos = st
            if us_pos not in env.SS_INDEX[contig]:
                continue
            if strand not in env.SS_INDEX[contig][us_pos]:
                continue
            if 'end' not in env.SS_INDEX[contig][us_pos][strand]:
                continue
            tmp_us_sites.append(st)

        tmp_ds_sites = []
        for en in range(min(ends)-10, max(ends)+10):
            ds_pos = en
            if ds_pos not in env.SS_INDEX[contig]:
                continue
            if strand not in env.SS_INDEX[contig][ds_pos]:
                continue
            if 'start' not in env.SS_INDEX[contig][ds_pos][strand]:
                continue
            tmp_ds_sites.append(en)

        if len(tmp_us_sites) == 0 and len(tmp_ds_sites) == 0:
            continue
        elif len(tmp_us_sites) > 0 and len(tmp_ds_sites) > 0:
            pass
        elif len(tmp_us_sites) > 0:
            pass
        elif len(tmp_ds_sites) > 0:
            pass
        else:
            continue


def recursive_splice_site(scores, ctg, strand):
    for st, en, scr in scores:
        if strand == '+' and (env.GENOME.seq(ctg, st-2, st) == 'AG' and env.GENOME.seq(ctg, st, st+2) == 'GT'):
            return st, en, scr
        if strand == '-' and (env.GENOME.seq(ctg, en, en+2) == 'CT' and env.GENOME.seq(ctg, en-2, en) == 'CA'):
            return st, en, scr
    return None, None, None


def curate_cirexons(circ, cluster):
    isoforms = {}
    starts = []
    ends = []
    for read in cluster:
        if read.cirexon == 'NA':
            continue

        exons = parse_cirexons(circ, read)
        if len(exons) == 0:
            continue

        for exon, exon_type in exons:
            if exon_type != '*-':
                starts.append(exon.start)
            if exon_type != '-*':
                ends.append(exon.end)

        if read.type == 'partial':
            continue
        isoforms[read.read_id] = [i[0] for i in exons]

    if len(isoforms) == 0:
        return None

    # Cluster junction sites
    tmp_starts = cluster_bins(starts, dis=10)
    tmp_ends = cluster_bins(ends, dis=10)

    convert_st = {}
    for tmp_st in tmp_starts:
        # circRNA start
        if circ.start in tmp_st:
            for i in tmp_st:
                convert_st[i] = circ.start

        # Canonical sites
        aval_st = []
        for i in set(tmp_st):
            i_ss = env.GENOME.seq(circ.contig, i-3, i-1)
            if circ.strand == '+' and i_ss == 'AG':
                aval_st.append(i)
            elif circ.strand == '-' and revcomp(i_ss) == 'GT':
                aval_st.append(i)
            else:
                pass

        tmp_st_counter = Counter(tmp_st)
        if aval_st:
            final_st = sorted(aval_st, key=lambda x: tmp_st_counter[x], reverse=True)[0]
        else:
            final_st = tmp_st_counter.most_common(n=1)[0][0]
        for i in tmp_st:
            convert_st[i] = final_st

    convert_en = {}
    for tmp_en in tmp_ends:
        # circRNA start
        if circ.end in tmp_en:
            for i in tmp_en:
                convert_en[i] = circ.end

        # Canonical sites
        aval_en = []
        for i in set(tmp_en):
            i_ss = env.GENOME.seq(circ.contig, i, i + 2)
            if circ.strand == '+' and i_ss == 'GT':
                aval_en.append(i)
            elif circ.strand == '-' and revcomp(i_ss) == 'AG':
                aval_en.append(i)
            else:
                pass

        tmp_en_counter = Counter(tmp_en)
        if aval_en:
            final_en = sorted(aval_en, key=lambda x: tmp_en_counter[x], reverse=True)[0]
        else:
            final_en = tmp_en_counter.most_common(n=1)[0][0]
        for i in tmp_en:
            convert_en[i] = final_en

    curated_exons = {}
    for read_id, exons in isoforms.items():
        tmp_exons = [Exon(convert_st[exon.start], convert_en[exon.end]) for exon in exons]

        while tmp_exons[0].end <= circ.start:
            tmp_exons = tmp_exons[1:]
            if len(tmp_exons) == 0:
                break
        if len(tmp_exons) == 0:
            continue

        while tmp_exons[-1].start >= circ.end:
            tmp_exons = tmp_exons[:-1]
            if len(tmp_exons) == 0:
                break
        if len(tmp_exons) == 0:
            continue

        tmp_exons = merge_cirexons(tmp_exons)
        if tmp_exons[0].start <= circ.start + 15 and tmp_exons[-1].end >= circ.end - 15:
            tmp_exons[0].start = circ.start
            tmp_exons[-1].end = circ.end
        else:
            continue

        curated_exons[read_id] = tmp_exons

    return curated_exons


def merge_cirexons(exons):
    if len(exons) == 1:
        return exons

    last_exon = exons[0]
    merged_exons = []
    for exon in exons[1:]:
        if exon.start <= last_exon.end + 10:
            last_exon = Exon(last_exon.start, exon.end)
        else:
            merged_exons.append(last_exon)
            last_exon = exon
    merged_exons.append(last_exon)

    return merged_exons


def curate_isoform(circ, curated_exons, cluster_res):
    final_isoforms = {}
    for tmp_seq, tmp_ids in cluster_res:
        tmp_isoform, tmp_len = merge_isoforms(circ, curated_exons, tmp_seq, tmp_ids)
        if tmp_isoform is None:
            continue

        if tmp_isoform in final_isoforms:
            final_isoforms[tmp_isoform][1] += tmp_ids
        else:
            final_isoforms[tmp_isoform] = [tmp_len, tmp_ids]
    if len(final_isoforms) == 0:
        return None, None

    total_cnt = sum([len(i[1]) for i in final_isoforms])
    ret = sorted(list(final_isoforms),
                 key=lambda x: (len(final_isoforms[x][1]), final_isoforms[x][0]),
                 reverse=True)
    major_len = final_isoforms[ret[0]][0]
    major_isoforms = [i for i in ret if len(final_isoforms[i][1]) >= 0.1 * total_cnt]
    return major_isoforms, major_len


def merge_isoforms(circ, curated_exons, seq, ids):
    from libs.striped_smith_waterman.ssw_wrap import Aligner
    aligner = Aligner(seq, match=10, mismatch=4, gap_open=8, gap_extend=2)

    tmp = [i for i in ids if i in curated_exons]
    exons = sorted(set([str(j) for i in tmp for j in curated_exons[i]]))
    if len(exons) == 0:
        return None, None

    exons = ['st', ] + exons + ['en']
    edges = np.zeros([len(exons), len(exons)])
    for i in tmp:
        tmp_exons = [str(j) for j in curated_exons[i]]
        edges[exons.index('st')][exons.index(tmp_exons[0])] += 1
        edges[exons.index(tmp_exons[-1])][exons.index('en')] += 1

        for l_exon, n_exon in pairwise(tmp_exons):
            edges[exons.index(l_exon)][exons.index(n_exon)] += 1

    cand_st, cand_en = np.where(edges == np.amax(edges))

    cand_score = [exon_score(circ, aligner, exons[i], exons[j]) for i, j in zip(cand_st, cand_en)]

    cand_idx = np.where(cand_score == np.amax(cand_score))[0][0]

    max_flow = []
    max_flow += iter_flow(circ, aligner, exons, edges, cand_st[cand_idx], -1)
    max_flow += iter_flow(circ, aligner, exons, edges, cand_en[cand_idx], 1)

    isoform = [exons[i] for i in max_flow]
    isoform_id = ','.join([i for i in isoform[1:-1]])
    isoform_len = sum([int(i.split('-')[1]) - int(i.split('-')[0]) for i in isoform[1:-1]])
    return isoform_id, isoform_len


def iter_flow(circ, aligner, exons, edges, coord, direction=-1):
    if coord == 0 or coord == edges.shape[0] - 1:
        return [coord, ]

    if direction == -1:
        max_l = np.where(edges[:, coord] == np.amax(edges[:, coord]))[0]
        max_score = [exon_score(circ, aligner, exons[i], exons[coord]) for i in max_l]
        max_idx = max_l[np.where(max_score == np.amax(max_score))[0][0]]
        return iter_flow(circ, aligner, exons, edges, max_idx, direction) + [coord, ]
    else:
        max_n = np.where(edges[coord] == np.amax(edges[coord]))[0]
        max_score = [exon_score(circ, aligner, exons[coord], exons[i]) for i in max_n]
        max_idx = max_n[np.where(max_score == np.amax(max_score))[0][0]]
        return [coord, ] + iter_flow(circ, aligner, exons, edges, max_idx, direction)


def exon_score(circ, aligner, l_exon, n_exon):
    query_seq = ''
    if l_exon != 'st':
        l_st, l_en = l_exon.split('-')
        tmp = env.GENOME.seq(circ.contig, int(l_st)-1, int(l_en))
        query_seq += tmp
    if n_exon != 'en':
        n_st, n_en = n_exon.split('-')
        tmp = env.GENOME.seq(circ.contig, int(n_st), int(n_en))
        query_seq += tmp
    if circ.strand == '-':
        query_seq = revcomp(query_seq)

    alignment = aligner.align(query_seq)
    return alignment.ref_end - alignment.ref_begin


def parse_cirexons(circ, read):
    exon_str = read.cirexon.split(',')
    exons = []
    for x in exon_str:
        st, en = x.split('|')[0].split('-')
        exons.append([Exon(st, en), x.split('|')[1]])
    return exons


def cluster_bins(pos, dis=10):
    clustered = []
    last_i = None
    for i in sorted(pos):
        if last_i is None:
            last_i = [i, ]
            continue
        if i > last_i[-1] + dis:
            clustered.append(last_i)
            last_i = [i, ]
        else:
            last_i.append(i)
    clustered.append(last_i)
    return clustered


def intron_splice_sites(contig, start, end, strand):
    us = env.GENOME.seq(contig, start, start + 2)
    ds = env.GENOME.seq(contig, end - 3, end - 1)
    if strand == '-':
        us, ds = revcomp(ds), revcomp(us)
    return signal_weight(us, ds)


def signal_weight(us, ds):
    if us == 'GT' and ds == 'AG':
        return 0
    else:
        return 1


def check_isoforms(circ, isoforms):
    concordance = []
    for iso_str in isoforms:
        exons = iso_str.split(',')
        if len(exons) == 1:
            concordance.append(True)
            continue

        introns = []
        for l_str, n_str in pairwise(exons):
            l_st, l_en = l_str.split('-')
            n_st, n_en = n_str.split('-')
            l_ss = env.GENOME.seq(circ.contig, int(l_en), int(l_en) + 2)
            n_ss = env.GENOME.seq(circ.contig, int(n_st) - 3, int(n_st) - 1)
            if circ.strand == '+' and l_ss == 'GT' and n_ss == 'AG':
                introns.append(1)
            elif circ.strand == '-' and revcomp(n_ss) == 'GT' and revcomp(l_ss) == 'AG':
                introns.append(1)
            else:
                introns.append(0)
        concordance.append(sum(introns) == len(introns))

    return sum(concordance) > 0


def correct_reads(reads_cluster, ref_fasta, gtf_index, intron_index, ss_index, threads):
    # Load reference genome
    genome = Fasta(ref_fasta)

    corrected_reads = []
    jobs = []
    pool = Pool(threads, env.initializer, (None, genome.contig_len, genome, gtf_index, intron_index, ss_index, ))

    for cluster in grouper(reads_cluster, 250):
        jobs.append(pool.apply_async(correct_chunk, (cluster, 200)))
    pool.close()

    prog = ProgressBar()
    prog.update(0)
    cnt = 0
    circ_num = defaultdict(int)
    for job in jobs:
        tmp_cluster, tmp_num = job.get()
        corrected_reads += tmp_cluster
        for i in tmp_num:
            circ_num[i] += tmp_num[i]
        cnt += 1
        prog.update(100 * cnt // len(jobs))
    pool.join()
    prog.update(100)

    return circ_num, corrected_reads


def circ_pos(x):
    ctg, pos = x.split(':')
    st, en = pos.split('-')
    return ctg, int(st), int(en)


def by_circ(x):
    ctg, pos = x.split(':')
    if ctg.startswith('chr'):
        ctg = ctg.lstrip('chr')

    try:
        idx = '{:02d}'.format(int(ctg))
    except ValueError as e:
        if ctg in ['X', 'x', 'Y', 'y']:
            idx = 'a'
        elif ctg in ['M', 'm']:
            idx = 'b'
        else:
            idx = 'c'

    st, en = pos.split('-')

    return idx, ctg, int(st), int(en)


def cal_exp_mtx(cand_reads, corrected_reads, ref_fasta, gtf_idx, out_dir, prefix):
    from collections import Counter

    genome = Fasta(ref_fasta)

    circ_reads = defaultdict(list)
    circ_info = {}
    reads_df = []

    # with open('{}/{}.fa'.format(out_dir, prefix), 'w') as fa:
    for reads, seqs, circ_id, strand, ss_id, us_free, ds_free, circ_len, isoforms in corrected_reads:
        # for tmp_id, tmp_seq in zip(reads, seqs):
        #     fa.write('>{}\t{}\t{}\n{}\n'.format(tmp_id, circ_id, strand, tmp_seq))

        # circRNA information
        ctg, st, en = circ_pos(circ_id)
        field = circ_attr(gtf_idx, ctg, st, en, strand)

        tmp_attr = 'circ_id "{}"; splice_site "{}"; equivalent_seq "{}"; circ_type "{}"; circ_len "{}";'.format(
            circ_id,
            ss_id,
            equivalent_seq(genome, ctg, st, en, strand),
            field['circ_type'] if field else 'Unknown',
            circ_len,
        )
        if isoforms:
            tmp_attr += ' isoform "{}";'.format('|'.join(isoforms))

        for key in 'gene_id', 'gene_name', 'gene_type':
            if key in field:
                tmp_attr += ' {} "{}";'.format(key, field[key])
        circ_info[circ_id] = [ctg, 'CIRI-long', 'circRNA', st, en, len(reads), strand, '.', tmp_attr, ]

        # Expression levels
        circ_reads[circ_id] += reads

        # Corrected reads
        for read_id in reads:
            read = cand_reads[read_id]
            tmp = [read_id, circ_id, read.circ_id, read.strand, read.cirexon,
                   read.ss, read.clip, read.segments, read.sample, read.type]
            reads_df.append(tmp)

    # circular reads
    reads_df = pd.DataFrame(
        reads_df, columns=['read_id', 'circ_id', 'tmp_id', 'strand', 'cirexons',
                           'signal', 'alignment', 'segments', 'sample', 'type'])
    reads_df.to_csv('{}/{}.reads'.format(out_dir, prefix), sep="\t", index=False)

    # circRNA information
    sorted_circ = sorted(list(circ_info), key=by_circ)
    with open('{}/{}.info'.format(out_dir, prefix), 'w') as out:
        for circ_id in sorted_circ:
            tmp_line = circ_info[circ_id]
            out.write('\t'.join([str(x) for x in tmp_line]) + '\n')

    # Expression level
    exp_df = {}
    for circ_id, reads in circ_reads.items():
        exp_df[circ_id] = Counter([cand_reads[i].sample for i in reads])
    exp_df = pd.DataFrame.from_dict(exp_df).transpose().fillna(0).reindex(sorted_circ)
    exp_df.to_csv('{}/{}.expression'.format(out_dir, prefix), sep="\t", index_label='circ_ID')

    return len(sorted_circ)


def equivalent_seq(genome, contig, start, end, strand):
    if strand is None:
        return 'Unknown'

    ds_seq = ''
    for i in range(100):
        if end + i > genome.contig_len[contig]:
            break
        if genome.seq(contig, start, start + i) == genome.seq(contig, end, end + i):
            ds_seq += genome.seq(contig, start, start + i)
        else:
            break

    us_seq = ''
    for j in range(100):
        if start - j < 0:
            break
        if genome.seq(contig, start - j, start) == genome.seq(contig, end - j, end):
            us_seq += genome.seq(contig, start - j, start)
        else:
            break

    tmp = us_seq[::-1] + ds_seq
    if strand == '+':
        return tmp
    else:
        return revcomp(tmp)


def circ_attr(gtf_index, ctg, start, end, strand):
    """
    annotate circRNA information
    """
    if ctg not in gtf_index:
        # LOGGER.warn('chrom of contig "{}" not in annotation gtf, please check'.format(ctg))
        return {}
    start_div, end_div = start // 500, end // 500

    host_gene = {}
    start_element = defaultdict(list)
    end_element = defaultdict(list)

    for x in range(start_div, end_div + 1):
        if x not in gtf_index[ctg]:
            continue
        for element in gtf_index[ctg][x]:
            # start site
            if element.start <= start <= element.end and (element.strand == strand or strand is None):
                start_element[element.type].append(element)
            # end site
            if element.start <= end <= element.end and (element.strand == strand or strand is None):
                end_element[element.type].append(element)
            # if element.type != 'gene':
            #     continue
            if element.end < start or end < element.start:
                continue
            if element.attr['gene_id'] not in host_gene:
                host_gene[element.attr['gene_id']] = element

    circ_type = {}
    forward_host_gene = []
    antisense_host_gene = []

    if len(host_gene) > 0:
        for gene_id in host_gene:
            if strand == 'None' or host_gene[gene_id].strand == strand:
                forward_host_gene.append(host_gene[gene_id])
                if 'exon' in start_element and 'exon' in end_element:
                    circ_type['exon'] = 1
                else:
                    circ_type['intron'] = 1
            else:
                antisense_host_gene.append(host_gene[gene_id])
                circ_type['antisense'] = 1
    else:
        circ_type['intergenic'] = 1

    if len(forward_host_gene) > 1:
        circ_type['gene_intergenic'] = 1

    field = {}
    if 'exon' in circ_type:
        field['circ_type'] = 'exon'
    elif 'intron' in circ_type:
        field['circ_type'] = 'intron'
    # elif 'gene_intergenic' in circ_type:
    #     field['circ_type'] = 'gene_intergenic'
    elif 'antisense' in circ_type:
        field['circ_type'] = 'antisense'
    else:
        field['circ_type'] = 'intergenic'

    if len(forward_host_gene) == 1:
        if 'gene_id' in forward_host_gene[0].attr:
            field['gene_id'] = forward_host_gene[0].attr['gene_id']
        if 'gene_name' in forward_host_gene[0].attr:
            field['gene_name'] = forward_host_gene[0].attr['gene_name']
        if 'gene_type' in forward_host_gene[0].attr:
            field['gene_type'] = forward_host_gene[0].attr['gene_type']
        elif 'gene_biotype' in forward_host_gene[0].attr:
            field['gene_type'] = forward_host_gene[0].attr['gene_biotype']
        else:
            pass
    elif len(forward_host_gene) > 1:
        tmp_gene_id = []
        tmp_gene_name = []
        tmp_gene_type = []
        for x in forward_host_gene:
            if 'gene_id' in x.attr:
                tmp_gene_id.append(x.attr['gene_id'])
            if 'gene_name' in x.attr:
                tmp_gene_name.append(x.attr['gene_name'])
            if 'gene_type' in x.attr:
                tmp_gene_type.append(x.attr['gene_type'])
            elif 'gene_biotype' in x.attr:
                tmp_gene_type.append(x.attr['gene_biotype'])
            else:
                pass
        if len(tmp_gene_id) > 0:
            field['gene_id'] = ','.join(tmp_gene_id)
        if len(tmp_gene_name) > 0:
            field['gene_name'] = ','.join(tmp_gene_name)
        if len(tmp_gene_type) > 0:
            field['gene_type'] = ','.join(tmp_gene_type)
    elif field['circ_type'] == 'antisense' and len(antisense_host_gene) > 0:
        tmp_gene_id = []
        tmp_gene_name = []
        tmp_gene_type = []
        for x in antisense_host_gene:
            if 'gene_id' in x.attr:
                tmp_gene_id.append(x.attr['gene_id'])
            if 'gene_name' in x.attr:
                tmp_gene_name.append(x.attr['gene_name'])
            if 'gene_type' in x.attr:
                tmp_gene_type.append(x.attr['gene_type'])
            elif 'gene_biotype' in x.attr:
                tmp_gene_type.append(x.attr['gene_biotype'])
            else:
                pass
        if len(tmp_gene_id) > 0:
            field['gene_id'] = ','.join(tmp_gene_id)
        if len(tmp_gene_name) > 0:
            field['gene_name'] = ','.join(tmp_gene_name)
        if len(tmp_gene_type) > 0:
            field['gene_type'] = ','.join(tmp_gene_type)
    else:
        pass

    return field



import os
import sys
import logging
from collections import namedtuple, defaultdict
LOGGER = logging.getLogger('CIRI-long')

import numpy as np


def collect_kmers(seq, k=8, use_hpc=False, is_circular=False):
    """
    split sequence into kmers
    :param seq: sequence
    :param k: k-mer size
    :param use_hpc: True if use homopolymer-compressed kmers
    :return: dict of kmers and occurence position
    """
    kmers = {}
    kmer_occ = defaultdict(list)

    split_func = split_hpc_kmers if use_hpc else split_kmers
    if is_circular:
        tmp_kmers, tmp_kmer_occ = collect_kmers(seq * 2, k, use_hpc, False)
        for i in tmp_kmers:
            if i < len(seq):
                continue
            kmers[i - len(seq)] = tmp_kmers[i]
            kmer_occ[tmp_kmers[i]].append(i - len(seq))
    else:
        for i, kmer in split_func(seq, k):
            kmers[i] = kmer
            kmer_occ[kmer].append(i)

    return kmers, kmer_occ


def all_kmers(k):
    import itertools
    all_bases = ['A', 'C', 'G', 'T']
    all_list = [all_bases, ] * k
    all_kmers = {''.join(kmer): idx for idx, kmer in enumerate(itertools.product(*all_list))}
    return all_kmers


def split_kmers(seq, k=8):
    for x in range(len(seq)):
        if x >= k - 1:
            yield x, seq[x-k+1:x+1]


def split_hpc_kmers(seq, k=8):
    hpc = [seq[0], ]
    for x, (i, j) in enumerate(zip(seq[:-1], seq[1:])):
        if i != j:
            hpc.append(j)
            if len(hpc) >= k:
                yield x + 1, ''.join(hpc[-k:])


def compress_seq(seq):
    hpc = [seq[0], ]
    for x, (i, j) in enumerate(zip(seq[:-1], seq[1:])):
        if i != j:
            hpc.append(j)
    return ''.join(hpc)


def estimate_distance(kmer_occ, p_indel, d_min):
    from scipy.stats.kde import gaussian_kde
    tuple_dis = []
    for k_occ in kmer_occ.values():
        for x1, x2 in zip(k_occ[:-1], k_occ[1:]):
            if x2 - x1 < d_min:
                continue
            tuple_dis.append(x2 - x1)
    if len(tuple_dis) <= 2:
        return None, None, 0

    dis_uniq = np.unique(tuple_dis)
    if len(dis_uniq) == 1:
        tuple_dis_mean = dis_uniq[0]
    else:
        # Kernel density estimation
        kernel = gaussian_kde(tuple_dis)
        tuple_dis_mean = dis_uniq[np.argmax(kernel(dis_uniq))]

    # Random walk model
    tuple_dis_delta = int(2.3 * np.sqrt(p_indel * tuple_dis_mean))
    tuple_dis_support = len([i for i in tuple_dis if abs(i - tuple_dis_mean) <= tuple_dis_delta])

    return tuple_dis_mean, tuple_dis_delta, tuple_dis_support


def optimal_hit(kmers, seed, s, e, use_hpc):
    """
    Return hit with smallest distance and nearest to optimal distance
    """
    from Levenshtein import distance
    from operator import itemgetter

    hits = []
    for i in range(s, e + 1):
        if i not in kmers:
            continue
        hits.append([i, kmers[i], distance(seed, kmers[i]), max(i - s, i - e)])

    # No hit
    if len(hits) == 0:
        return None, None, len(seed), None

    # For HPC hits
    if use_hpc and len(hits) > 1:
        hits = collapse_hits(hits)

    return sorted(hits, key=itemgetter(2, 3))[0]


def collapse_hits(hits):
    """
    For hits with same HPC k-mer content and distance, collapse into the last base
    """
    collapsed = []
    for i, j in zip(hits[:-1], hits[1:]):
        if j[0] == i[0] + 1 and j[1] == i[1]:
            continue
        else:
            collapsed.append(j)
    return collapsed


def circular_hits(kmers, k, tuple_dis_mean, tuple_dis_delta, p_match, use_hpc):
    pos = sorted(list(kmers))
    hits = {}
    for i in pos:
        j, _, score, _ = optimal_hit(kmers, kmers[i],
                                     i + tuple_dis_mean - tuple_dis_delta,
                                     i + tuple_dis_mean + tuple_dis_delta,
                                     use_hpc)
        if score > int(k * (1 - p_match)):
            continue
        hits[i] = (i, j, j - i, score)
    return hits


def optimal_chains(hits):
    """
    Intersect hits into chains
    """
    pos = sorted(list(hits))
    s0, e0, _, score0 = hits[pos[0]]

    chains = [[(s0, e0, score0), ], ]
    for i in pos[1:]:
        s, e, _, score = hits[i]
        if chains[-1][-1][0] < s < chains[-1][-1][1] <= e:
            chains[-1].append((s, e, score))
        elif chains[-1][-1][1] < s:
            chains.append([(s, e, score), ])
        else:
            pass

    # Sort chains by length
    chains = sorted(chains, key=lambda x: x[-1][1] - x[0][0], reverse=True)

    return chains[0]


def split_sequence(primary_chain, seq, p_match, p_indel):
    from Bio import pairwise2

    match = 10
    mismatch = -4
    gap = -8
    extension = -2

    final = primary_chain[-1][1]

    for s, e, score in primary_chain:
        if score == 0:
            break
    boundaries = [s, e, ]

    idx = 0

    hits = {hit[0]: hit for hit in primary_chain}
    hit_pos = [hit[0] for hit in primary_chain]

    last_x = e
    while last_x < final:
        if last_x in hits:
            idx = hit_pos.index(last_x)
            last_x = hits[last_x][1]
            boundaries.append(last_x)
        else:
            # Find nearest hits
            while hit_pos[idx] < last_x:
                if idx >= len(hit_pos) - 1 or hit_pos[idx + 1] >= last_x:
                    break
                idx += 1
            if idx >= len(hit_pos) - 1:
                break

            # local alignment
            q_s, r_s, _ = hits[hit_pos[idx]]
            q_e, r_e, _ = hits[hit_pos[idx + 1]]

            alignments = pairwise2.align.globalms(seq[q_s:q_e + 1], seq[r_s:r_e + 1],
                                                  match, mismatch, gap, extension)

            q_shift, r_shift = 0, 0
            q_seq, r_seq, score, _, _ = alignments[0]
            for i, j in zip(q_seq, r_seq):
                q_shift += int(i in 'atcgATCG')
                r_shift += int(j in 'atcgATCG')
                if q_shift == last_x - q_s:
                    break
            # TODO: Stop searching when the score is too slow
            last_x = r_s + r_shift
            boundaries.append(last_x)

    segments = [(i, j) for i, j in zip(boundaries[:-1], boundaries[1:])]
    if segments[-1][-1] < final:
        segments.append((segments[-1][-1], final))

    return segments


def circular_finder(read_id, seq, k=8, use_hpc=True, p_match=.85, p_indel=.1, d_min=40, support_min=2):
    assert 0 <= p_indel <= 1
    assert 0 <= p_match <= 1

    # Split k-mers
    kmers, kmer_occ = collect_kmers(seq, k, use_hpc)

    # Approximately distance of circular segments
    tuple_dis_mean, tuple_dis_delta, tuple_dis_support = estimate_distance(kmer_occ, p_indel, d_min)
    if tuple_dis_support < support_min:
        return None, None

    # Hits of kmers
    tuple_hits = circular_hits(kmers, k, tuple_dis_mean, tuple_dis_delta, p_match, use_hpc)
    if not tuple_hits:
        return None, None

    # Primary chain
    primary_chain = optimal_chains(tuple_hits)

    # Split into segments
    segments = split_sequence(primary_chain, seq, p_match, p_indel)

    is_circular = 1
    # Head and tail position
    if segments[0][0] > 100 or segments[-1][-1] < len(seq) - 100:
        is_circular = 0

    if len(segments) < 2:
        is_circular = 0

    if len(segments) == 2 and segments[-1][1] - segments[-1][0] < 30:
        is_circular = 0

    return segments, is_circular


def find_consensus(header, seq):
    from spoa import poa
    from Levenshtein import distance

    # if header not in 'ENSMUST00000021181|ENSMUSG00000020831|11:70236877-70237625|-|197_1126_aligned_43558_F_52_821_60':
    #     return None, None, None

    # Trim sequence
    if len(seq) <= 50:
        return None, None, None

    # Repeat segments
    chains = None
    is_circular = 0
    for k, is_hpc in [(11, False), (8, False), (11, True), (8, True)]:
        tmp_chain, tmp_circular = circular_finder(header, seq, k=k, use_hpc=is_hpc)
        if tmp_chain is None:
            continue
        if tmp_circular == 1:
            chains = tmp_chain
            is_circular = 1
            break
        elif chains is None:
            chains = tmp_chain
        else:
            pass
    if chains is None:
        return None, None, None

    # Chains
    if len(chains) < 2:
        return None, None, None

    fasta = [seq[s:e] for s, e in chains]
    ccs, _ = poa(fasta, 2, False, 10, -4, -8, -2, -24, -1)

    # Check segment similarity
    tail = fasta[-1]
    if len(fasta) == 2:
        dis_body = distance(fasta[0][:len(tail)], ccs[:len(tail)]) / len(tail)
    else:
        dis_body = max([distance(i, ccs) / len(ccs) for i in fasta[:-1]])
    dis_tail = distance(tail, ccs[:len(tail)]) / len(tail)

    if dis_body > 0.2 or dis_tail > 0.35:
        return None, None, None

    segments = ';'.join(['{}-{}'.format(s, e) for s, e in chains])

    return segments, ccs, is_circular


def worker(chunk):
    ret = []
    for header, seq in chunk:
        segments, ccs, is_circular = find_consensus(header, seq)
        if segments is None or ccs is None or is_circular == 0:
            continue
        ret.append((header, seq, segments, ccs))
    return len(chunk), ret


def find_ccs_reads(in_file, out_dir, prefix, threads, debugging):
    import gzip
    from multiprocessing import Pool
    from CIRI_long.utils import to_str
    from CIRI_long.logger import ProgressBar
    pool = Pool(threads)
    jobs = []

    # Auto detect input format
    is_fastq = 1
    is_gz = 1
    if in_file.endswith('.fa') or in_file.endswith('.fasta'):
        is_fastq = 0
        is_gz = 0
        fq = open(in_file, 'r')
    elif in_file.endswith('.fa.gz') or in_file.endswith('.fasta.gz'):
        is_fastq = 0
        is_gz = 1
        fq = gzip.open(in_file, 'rb')
    elif in_file.endswith('.fq') or in_file.endswith('.fastq'):
        is_gz = 0
        fq = open(in_file, 'r')
    elif in_file.endswith('.fq.gz') or in_file.endswith('.fastq.gz'):
        fq = gzip.open(in_file, 'rb')
    else:
        sys.exit('Wrong format of input')

    chunk = []
    chunk_size = 250
    chunk_cnt = 0
    for line in fq:
        if is_gz:
            header = to_str(line).rstrip().split(' ')[0]
            seq = to_str(fq.readline()).rstrip()
        else:
            header = line.rstrip().split(' ')[0]
            seq = fq.readline().rstrip()

        if is_fastq:
            header = header.lstrip('@')
            fq.readline()
            fq.readline()
        else:
            header = header.lstrip('>')

        # Split into chunks
        chunk.append((header, seq))
        if len(chunk) == chunk_size:
            jobs.append(pool.apply_async(worker, (chunk, )))
            chunk = []
            chunk_cnt += 1

    if len(chunk) > 0:
        jobs.append(pool.apply_async(worker, (chunk, )))
        chunk_cnt += 1
    pool.close()
    fq.close()

    prog = ProgressBar()
    prog.update(0)
    finished_chunk = 0

    total_reads = 0
    ro_reads = 0

    ccs_seq = {}
    with open('{}/tmp/{}.ccs.fa'.format(out_dir, prefix), 'w') as out, \
            open('{}/tmp/{}.raw.fa'.format(out_dir, prefix), 'w') as trimmed:
        for job in jobs:
            tmp_cnt, ret = job.get()
            total_reads += tmp_cnt
            for header, seq, segments, ccs in ret:
                ro_reads += 1
                out.write('>{}\t{}\t{}\n{}\n'.format(header, segments, len(ccs), to_str(ccs)))
                trimmed.write('>{}\n{}\n'.format(header, seq))
                ccs_seq[header] = [segments, to_str(ccs), seq]

            finished_chunk += 1
            prog.update(100 * finished_chunk // chunk_cnt)
    prog.update(100)
    pool.join()

    return total_reads, ro_reads, ccs_seq


def load_ccs_reads(out_dir, prefix):
    ccs_seq = {}
    with open('{}/tmp/{}.ccs.fa'.format(out_dir, prefix), 'r') as f:
        for line in f:
            header = line.rstrip()
            content = header.split('\t')
            seq = f.readline().rstrip()
            ccs_seq[content[0].lstrip('>')] = [content[1], seq]

    with open('{}/tmp/{}.raw.fa'.format(out_dir, prefix), 'r') as f:
        for line in f:
            header = line.rstrip().split('\t')[0].lstrip('>')
            seq = f.readline().rstrip()
            ccs_seq[header].append(seq)
    return ccs_seq



#! /usr/bin/env python
# -*- encoding:utf-8 -*-
import os
import sys
import time
import logging
from multiprocessing.managers import BaseManager


class ProgressBar(object):
    def __init__(self, width=50):
        self.last_x = -1
        self.width = width

    def update(self, x):
        assert 0 <= x <= 100
        if self.last_x == int(x):
            return
        self.last_x = int(x)
        p = int(self.width * (x / 100.0))
        time_stamp = time.strftime("[%a %Y-%m-%d %H:%M:%S]", time.localtime())
        sys.stderr.write('\r%s [%-5s] [%s]' % (time_stamp, str(int(x)) + '%', '#' * p + '.' * (self.width - p)))
        sys.stderr.flush()
        if x == 100:
            sys.stderr.write('\n')

    def self_update(self):
        self.update(max(100, self.last_x + 1))


class ProgManager(BaseManager):
    pass


ProgManager.register('Prog', ProgressBar)


def get_logger(logger_name='logger', fname=None, verbosity=False):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    level = logging.DEBUG if verbosity else logging.INFO

    # LOG format
    fmt = "%(asctime)-15s [%(levelname)-5s] %(message)s"
    datefmt = "[%a %Y-%m-%d %H:%M:%S]"
    formatter = logging.Formatter(fmt, datefmt)

    # LOG file
    if fname is not None:
        fh = open(fname, 'w')
        fh.close()
        file_handler = logging.FileHandler(fname)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    # LOG console
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setLevel(level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    return logger


def find_logger_basefilename(logger):
    log_file = None
    parent = logger.__dict__['parent']
    if parent.__class__.__name__ == 'RootLogger':
        for h in logger.__dict__['handlers']:
            if h.__class__.__name__ == 'FileHandler':
                log_file = h.baseFilename
    else:
        log_file = find_logger_basefilename(parent)
    return log_file






#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
@package pyssw
@brief Python standalone program for ssw alignment using the C library
Complete-Striped-Smith-Waterman-Library
Biopython module is require for fastq/fastq parsing
@copyright  [The MIT licence](http://opensource.org/licenses/MIT)
@author     Adrien Leger - 2014
* <adrien.leger@gmail.com>
* <adrien.leger@inserm.fr>
* <adrien.leger@univ-nantes.fr>
* [Github](https://github.com/a-slide)
* [Atlantic Gene Therapies - INSERM 1089] (http://www.atlantic-gene-therapies.fr/)
"""

#~~~~~~~GLOBAL IMPORTS~~~~~~~#
# Standard library packages
import optparse
import sys
from time import time
import gzip

#~~~~~~~MAIN FUNCTION~~~~~~~#
def main (opt):

    print ("Inport subject sequence")
    # Import fasta subject
    if opt.subject.rpartition(".")[2].lower() == "gz":
        subject_handle = gzip.open(opt.subject, "r")
    else:
        subject_handle = open(opt.subject, "r")
    subject = SeqIO.read(subject_handle, "fasta")

    print ("Inport query sequences and count the number of sequences")
    # Import fasta subject
    if opt.query.rpartition(".")[2].lower() == "gz":
        nseq = count_seq(opt.query, opt.qtype, True)
        query_handle = gzip.open(opt.query, "r")
    else:
        nseq = count_seq(opt.query, opt.qtype, False)
        query_handle = open(opt.query, "r")
    query_gen = SeqIO.parse(query_handle, opt.qtype)

    print("{} contains {} sequences to align".format(opt.query, nseq))
    # Calculate a step list for the progress bar
    nseq_list = [int(nseq*i/100.0) for i in range(5,101,5)]

    print ("Initialize ssw aligner with the subject sequence")
    # Init the an Aligner object with the reference value
    ssw = Aligner(
        str(subject.seq),
        match=int(opt.match),
        mismatch=int(opt.mismatch),
        gap_open=int(opt.gap_open),
        gap_extend= int(opt.gap_extend),
        report_secondary=False,
        report_cigar=True)

    # Write the header of the SAM file
    with open("result.sam", "w") as f:
        f.write("@HD\tVN:1.0\tSO:unsorted\n")
        f.write("@SQ\tSN:{}\tLN:{}\n".format(subject.id, len(subject.seq)))
        f.write("@PG\tID:Striped-Smith-Waterman\tPN:pyssw\tVN:0.1\n")
        f.write("@CO\tScore_values = match {}, mismatch {}, gap_open {}, gap_extend {}\n".format(
            opt.match,
            opt.mismatch,
            opt.gap_open,
            opt.gap_extend))
        f.write("@CO\tFilter Options = min_score {}, min_len {}\n".format(
            opt.min_score,
            opt.min_len))

        print ("Starting alignment of queries against the subject sequence")
        start = time()
        # Align each query along the subject an write result in a SAM file
        i = 0
        for query in query_gen:

            # Find the best alignment
            if opt.reverse:
                al, orient = find_best_align (ssw, query, float(opt.min_score), int(opt.min_len))
            else:
                al, orient = ssw.align(str(query.seq), float(opt.min_score), int(opt.min_len)), True

            # If valid match found
            if al:
                f.write(sam_line(
                    qname=query.id,
                    flag=0 if orient else 16,
                    rname=subject.id,
                    pos=al.ref_begin+1,
                    cigar=al.cigar_string,
                    seq=str(query.seq),
                    qual=SeqIO.QualityIO._get_sanger_quality_str(query) if opt.qtype == "fastq" else "*",
                    tags=["AS:i:{}".format(al.score)]))

            # If no valid match found and -u flag activated (report unaligned)
            elif opt.unaligned:
                f.write(sam_line(
                    qname=query.id,
                    flag=4,
                    seq=str(query.seq),
                    qual=SeqIO.QualityIO._get_sanger_quality_str(query) if opt.qtype == "fastq" else "*"))
            # Else = match unreported

            # Progress bar
            i+=1
            if i in nseq_list:
                frac = i/float(nseq)
                t = time()-start
                print ("{} sequences \t{}% \tRemaining time = {}s".format(i, int(frac*100), round(t/frac-t, 2)))

        print ("\n{} Sequences processed in {}s".format(i, round(time()-start, 2)))

#~~~~~~~HELPER FUNCTIONS~~~~~~~#


def sam_line (qname='*', flag=4, rname='*', pos=0, mapq=0, cigar='*', rnext='*', pnext=0, tlen=0, seq='*', qual='*', tags=None):
    """
    Return a minimal sam line = by default return an undetermined sam line. Check the document
    [SAM Format Specification](http://samtools.sourceforge.net/SAM1.pdf) for a full description.
    @param qname Query template NAME
    @param flag bitwise FLAG
    @param rname Reference sequence NAME of the alignment
    @param pos 1-based leftmost mapping POSition of the first matching base
    @param mapq MAPping Quality
    @param cigar CIGAR string
    @param rnext Reference sequence name of the primary alignment of the mate
    @param pnext 1-based leftmost position of the primary alignment of the mate
    @param tlen signed observed Template LENgth
    @param seq segment SEQuence
    @param qual ASCII of base QUALity plus 33
    @param tags list of optional tags
    @return A Sam alignment line
    """
    if tags:
        return "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(
            qname, flag, rname, pos, mapq, cigar, rnext, pnext, tlen, seq, qual, " ".join(tags))
    else:
        return "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}".format(
            qname, flag, rname, pos, mapq, cigar, rnext, pnext, tlen, seq, qual)

def find_best_align (ssw, query, min_score, min_len):

    # Align reverse and forward query
    forward_al = ssw.align(str(query.seq), min_score, min_len)
    reverse_al = ssw.align(str(query.seq.reverse_complement()), min_score, min_len)

    # Decision tree to return the best aligned sequence taking into acount the absence of result
    # by ssw_wrap in case of score filtering

    if not forward_al:
        if not reverse_al:
            return (None, None)
        else:
            return (reverse_al, False)

    else:
        if not reverse_al:
            return (forward_al, True)
        else:
            if forward_al.score >= reverse_al.score:
                return (forward_al, True)
            else:
                return (reverse_al, False)

def count_seq (filename, seq_type="fasta", gziped=False):
    """
    Count the number of sequences in a fastq or a fastq file
    @param filename Path to a valid readeable file
    @param file_type Should be either fastq or fastq. Default fasta
    @param gziped Boolean indicating if the file is gziped or not. Default False
    """
    #Standard library import
    import gzip
    from mmap import mmap

    # Verify if the file is fasta or fastq type
    assert seq_type in ["fasta", "fastq"], "The file has to be either fastq or fasta format"

    # Open the file
    if gziped:
        f = gzip.open(filename, "r")
    else:
        f = open(filename, "r")

    # FASTA Find a start line seq character ">" an increment the counter each time
    if seq_type ==  "fasta":
        nline = 0
        for line in f:
            if line[0] == ">":
                nline+=1
        f.close()
        return nline

    # FASTQ No motif to find, but 4 lines correspond to 1 sequence
    else:
        nline = 0
        for line in f:
            nline+=1
        f.close()
        return nline/4

def optparser():

    print("Parse command line options")
    # Usage and version strings
    program_name = "pyssw"
    program_version = 0.1
    version_string = "{}\t{}".format(program_name, program_version)
    usage_string = "{}.py -s subject.fasta -q fastq (or fasta) [Facultative options]".format(program_name)
    optparser = optparse.OptionParser(usage = usage_string, version = version_string)

    # Define optparser options
    hstr = "Path of the fasta file containing the subject genome sequence. Can be gziped. [REQUIRED] "
    optparser.add_option( '-s', '--subject', dest="subject", help=hstr)
    hstr = "Path of the fastq or fasta file containing the short read to be aligned. Can be gziped. [REQUIRED]"
    optparser.add_option( '-q', '--query', dest="query", help=hstr)
    hstr = "Type of the query file = fastq or fasta. [default: fastq]"
    optparser.add_option( '-t', '--qtype', dest="qtype", default="fastq", help=hstr)
    hstr = "Positive integer for weight match in genome sequence alignment. [default: 2]"
    optparser.add_option( '-m', '--match', dest="match",default=2, help=hstr)
    hstr = "Positive integer. The negative value will be used as weight mismatch in genome sequence alignment. [default: 2]"
    optparser.add_option( '-x', '--mismatch', dest="mismatch", default=2, help=hstr)
    hstr = "Positive integer. The negative value will be used as weight for the gap opening. [default: 3]"
    optparser.add_option( '-o', '--gap_open', dest="gap_open", default=3, help=hstr)
    hstr = "Positive integer. The negative value will be used as weight for the gap opening. [default: 1]"
    optparser.add_option( '-e', '--gap_extend', dest="gap_extend", default=1, help=hstr)
    hstr = "Integer. Consider alignments having a score <= as not aligned. [default: 0]"
    optparser.add_option( '-f', '--min_score', dest="min_score", default=0, help=hstr)
    hstr = "Integer. Consider alignments having a length <= as not aligned. [default: 0]"
    optparser.add_option( '-l', '--min_len', dest="min_len", default=0, help=hstr)
    hstr = "Flag. Align query in forward and reverse orientation and choose the best alignment. [Set by default]"
    optparser.add_option( '-r', '--reverse', dest="reverse", action="store_true", default=True, help=hstr)
    hstr = "Flag. Write unaligned reads in sam output [Unset by default]"
    optparser.add_option( '-u', '--unaligned', dest="unaligned", action="store_true", default=False, help=hstr)

    # Parse arg and return a dictionnary_like object of options
    opt, args = optparser.parse_args()

    if not opt.subject:
        print ("\nERROR: a subject fasta file has to be provided (-s option)\n")
        optparser.print_help()
        sys.exit()

    if not opt.query:
        print ("\nERROR: a query fasta or fastq file has to be provided (-q option)\n")
        optparser.print_help()
        sys.exit()

    return opt

#~~~~~~~TOP LEVEL INSTRUCTIONS~~~~~~~#

if __name__ == '__main__':

    # try to import Third party and local packages
    try:
        from Bio import SeqIO
    except ImportError:
        print ("ERROR: Please install Biopython package")
        sys.exit()

    try:
        from ssw_wrap import Aligner
    except ImportError:
        print ("ERROR: Please place ssw_wrap in the current directory or add its dir to python path")
        sys.exit()

    # Parse command line arguments
    opt = optparser()
    # Run the main function
    main(opt)



import os
import sys
import re
import itertools
import threading
import _thread as thread


def quit_function(fn_name):
    sys.stderr.write('{0} took too long\n'.format(fn_name))
    sys.stderr.flush()
    thread.interrupt_main()


def exit_after(s):
    """
    use as decorator to exit process if
    function takes longer than s seconds
    """
    def outer(fn):
        def inner(*args, **kwargs):
            timer = threading.Timer(s, quit_function, args=[fn.__name__])
            timer.start()
            try:
                result = fn(*args, **kwargs)
            finally:
                timer.cancel()
            return result
        return inner
    return outer


def check_file(file_name):
    if os.path.exists(file_name) and os.path.isfile(file_name):
        return os.path.abspath(file_name)
    else:
        sys.exit('File: {}, not found'.format(file_name))


def check_dir(dir_name):
    if os.path.exists(dir_name):
        if os.path.isdir(dir_name):
            # Output directory already exists
            pass
        else:
            sys.exit('Directory: {}, clashed with existed files'.format(dir_name))
    else:
        os.mkdir(dir_name)
    return os.path.abspath(dir_name)


def to_str(bytes_or_str):
    # Return Instance of str
    if isinstance(bytes_or_str, bytes):
        value = bytes_or_str.decode('utf-8')
    else:
        value = bytes_or_str
    return value


def to_bytes(bytes_or_str):
    # Return Instance of bytes
    if isinstance(bytes_or_str, str):
        value = bytes_or_str.encode('utf-8')
    else:
        value = bytes_or_str
    return value


def empty_iter(iterable):
    try:
        first = next(iterable)
    except StopIteration:
        return None
    return itertools.chain([first], iterable)


def grouper(iterable, n, fillvalue=None):
    from itertools import zip_longest
    """
    Collect data info fixed-length chunks or blocks
    grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    """

    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=None)


def pairwise(iterable):
    from itertools import tee
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def tree():
    from collections import defaultdict
    return defaultdict(tree)


def flatten(x):
    """
    Flatten list of lists
    """
    import itertools

    flatted_list = list(itertools.chain(*x))
    return flatted_list


def min_sorted_items(iters, key, reverse=False):
    from operator import itemgetter
    x = sorted(iters, key=itemgetter(key), reverse=reverse)
    return [i for i in x if i[key] == x[0][key]]


def revcomp(seq):
    trantab = str.maketrans("ATCG", "TAGC")
    return seq.translate(trantab)[::-1]


def transform_seq(seq, bsj):
    return seq[bsj:] + seq[:bsj]


def get_junc_seq(seq, bsj, width=25):
    st, en = bsj - width, bsj + width
    if len(seq) <= 2 * width:
        return seq[bsj-len(seq)//2:] + seq[:bsj-len(seq)//2]

    if st < 0:
        if en < 0:
            return seq[st:en]
        else:
            return seq[st:] + seq[:en]
    elif en > len(seq):
        return seq[st:] + seq[:en-len(seq)]
    else:
        return seq[st:en]


def shift_base(seq):
    base = 0
    for i in seq:
        if i in 'atcgATGC':
            break
        else:
            base += 1
    return base



import logging
import subprocess
LOGGER = logging.getLogger('CIRI-long')


def run_ccs(in_file, out_dir, prefix, threads, debugging):
    ccs_fa = '{}/tmp/{}.ccs.fa'.format(out_dir, prefix)
    raw_fa = '{}/tmp/{}.raw.fa'.format(out_dir, prefix)

    ccs_cmd = 'ccs -i {} -o {} -r {} -t {}'.format(
        in_file, ccs_fa, raw_fa, threads,
    )
    with open('{}/{}.log'.format(out_dir, prefix), 'a') as log:
        LOGGER.debug(ccs_cmd)
        ret = subprocess.call(ccs_cmd, shell=True, stdout=log)

    return ret



import os
import sys
import re
import logging
from collections import defaultdict

import pysam
from CIRI_long import env
from CIRI_long.utils import *

LOGGER = logging.getLogger('CIRI-long')

OPERATION = {
    'M': 0,
    'I': 1,
    'D': 2,
    'N': 3,
    'S': 4,
    'H': 5,
    'P': 6,
    '=': 7,
    'X': 8,
    0: 'M',
    1: 'I',
    2: 'D',
    3: 'N',
    4: 'S',
    5: 'H',
    6: 'P',
    7: '=',
    9: 'X',
}

SPLICE_SIGNAL = {
    ('GT', 'AG'): 0,  # U2-type
    ('GC', 'AG'): 1,  # U2-type
    ('AT', 'AC'): 2,  # U12-type
    ('GT', 'AC'): 2,  # U12-type
    ('AT', 'AG'): 2,  # U12-type
    # ('GT', 'TG'): 3,  # non-canonical
    # ('AT', 'AG'): 3,  # non-canonical
    # ('GA', 'AG'): 3,  # non-canonical
    # ('GG', 'AG'): 3,  # non-canonical
    # ('GT', 'GG'): 3,  # non-canonical
    # ('GT', 'AT'): 4,  # non-canonical
    # ('GT', 'AA'): 4,  # non-canonical
}


class GTFParser(object):
    """
    Class for parsing annotation gtf
    """

    def __init__(self, content):
        self.contig = content[0]
        self.source = content[1]
        self.type = content[2]
        self.start, self.end = int(content[3]), int(content[4])
        self.strand = content[6]
        self.attr_string = content[8]

    @property
    def attr(self):
        """
        Parsing attribute column in gtf file
        """
        import re
        field = {}
        for attr_values in [re.split(r'\s+', i.strip()) for i in self.attr_string.split(';')[:-1]]:
            key, value = attr_values[0], attr_values[1:]
            field[key] = ' '.join(value).strip('"')
        return field


class Hit(object):
    """
    API for bwapy alignment
    Alignment(rname='chr7', orient='-', pos=131329219, mapq=60, cigar='51S37M', NM=0)
    """

    def __init__(self, aln):
        self.ctg = aln.rname
        self.strand = 1 if aln.orient == '+' else -1
        self.cigar_string = aln.cigar
        self.r_st = aln.pos
        self.r_en, self.q_st, self.q_en = self.__parse_cigar()
        self.mlen = self.q_en - self.q_st
        self.is_primary = 0

    @property
    def cigar(self):
        return [(int(length), OPERATION[operation]) for length, operation in
                re.findall(r'(\d+)([MIDNSHP=X])', self.cigar_string)]

    def __parse_cigar(self):
        r_en = self.r_st
        q_st, q_en = 0, 0
        for length, operation in self.cigar:
            if operation == 0:
                q_en += length
                r_en += length
            elif operation == 1:
                q_en += length
            elif operation in [2, 3]:
                r_en += length
            elif operation in [4, 5]:
                if q_st == 0:
                    q_st = length
                    q_en = length
            else:
                pass
        return r_en, q_st, q_en

    def __str__(self):
        return '\t'.join([str(x) for x in [self.q_st, self.q_en, self.ctg, self.r_st, self.r_en, self.mlen, self.blen,
                                           self.cigar_string]])


class SubHit(object):
    def __init__(self, hit, r_st, q_st, cigar):
        self.ctg = hit.ctg
        self.strand = hit.strand
        self.cigar = cigar
        self.r_st = r_st
        self.r_en, self.q_st, self.q_en = self.__parse_cigar(q_st)
        self.mlen, self.blen = self.__match()
        self.is_primary = 0

    def __parse_cigar(self, q_st):
        r_en = self.r_st
        q_en = q_st
        for length, operation in self.cigar:
            if operation == 0:
                q_en += length
                r_en += length
            elif operation == 1:
                q_en += length
            elif operation in [2, 3]:
                r_en += length
            elif operation in [4, 5]:
                if q_st == 0:
                    q_st += length
                    q_en += length
            else:
                pass
        return r_en, q_st, q_en

    def __match(self):
        mlen, blen = 0, 0
        for l, o in self.cigar:
            if o in [0, 1]:
                mlen += l
            if o in [0, 1, 2]:
                blen += l
        return mlen, blen

    @property
    def cigar_string(self):
        return ''.join(['{}{}'.format(length, OPERATION[operation]) for length, operation in self.cigar])

    def __str__(self):
        return '\t'.join([str(x) for x in [self.q_st, self.q_en, self.ctg, self.r_st, self.r_en, self.mlen, self.blen,
                                           self.cigar_string]])


class Aligner(object):
    """
    API for unify bwapy to mappy::Aligner
    """
    def __init__(self, aligner):
        self.aligner = aligner

    def map(self, seq):
        alignments = self.aligner.align_seq(seq)
        if alignments:
            hits = [Hit(aln) for aln in alignments]
            hits[0].is_primary = 1
            return hits
        else:
            return None


class Faidx(object):
    """
    API for unify pysam::Fastafile and mappy2::Aligner
    """
    def __init__(self, infile):
        if not os.path.exists(infile + '.fai'):
            LOGGER.warn('Index of reference genome not found, generating ...')
        try:
            faidx = pysam.FastaFile(infile)
        except ValueError:
            LOGGER.error('Cannot generate index of reference genome, index file is missing')
            sys.exit()
        except IOError:
            LOGGER.error('Cannot generate index of reference genome, file could not be opened')
            sys.exit()

        self.faidx = faidx
        self.contig_len = {contig: faidx.get_reference_length(contig) for contig in faidx.references}

    def seq(self, contig, start, end):
        return self.faidx.fetch(contig, start, end)

    def close(self):
        self.faidx.close()


class Fasta(object):
    """
    load fasta file into memory
    """
    def __init__(self, infile):
        faidx = Faidx(infile)
        self.contig_len = faidx.contig_len
        self.genome = {ctg: faidx.seq(ctg, 0, size) for ctg, size in self.contig_len.items()}
        faidx.close()

    def seq(self, contig, start, end):
        if contig not in self.genome:
            return None
        return self.genome[contig][start:end]


def index_annotation(gtf):
    """
    Generate binned index for element in gtf
    """
    from CIRI_long.utils import tree

    LOGGER.info('Loading annotation gtf ..')
    gtf_index = defaultdict(dict)
    intron_index = defaultdict(dict)
    splice_site_index = tree()

    last_exon = None
    with open(gtf, 'r') as f:
        for line in f:
            if line.startswith('#'):
                continue
            content = line.rstrip().split('\t')
            # only include gene and exon feature for now
            if content[2] not in ['gene', 'exon']:
                continue

            parser = GTFParser(content)

            # Extract splice site
            if content[2] == 'exon':
                splice_site_index[parser.contig][parser.start][parser.strand]['start'] = 1
                splice_site_index[parser.contig][parser.end][parser.strand]['end'] = 1

                # Load intron
                if last_exon is not None and last_exon.attr['transcript_id'] == parser.attr['transcript_id']:
                    intron_start = last_exon.end if last_exon.strand == '+' else last_exon.start
                    intron_end = parser.start if parser.strand == '+' else parser.end
                    intron_strand = parser.strand

                    intron_start, intron_end = min(intron_start, intron_end), max(intron_start, intron_end)
                    start_div, end_div = intron_start // 500, intron_end // 500
                    for i in range(start_div, end_div + 1):
                        intron_index[parser.contig].setdefault(i, []).append((intron_start, intron_end, intron_strand))

                last_exon = parser

            # Binned index
            start_div, end_div = parser.start // 500, parser.end // 500
            for i in range(start_div, end_div + 1):
                gtf_index[parser.contig].setdefault(i, []).append(parser)

    return gtf_index, intron_index, splice_site_index


def index_circ(circ_file, circ_ss_idx):
    """
    Generate binned index for element in gtf
    """
    from CIRI_long.utils import tree
    from pathlib import Path

    circ_path = Path(circ_file)
    if circ_ss_idx is None:
        circ_ss_idx = tree()

    if circ_path.suffix == ".gtf":
        LOGGER.info('Loading additional circRNA gtf ..')
        with open(circ_path, 'r') as f:
            for line in f:
                if line.startswith('#'):
                    continue
                content = line.rstrip().split('\t')
                parser = GTFParser(content)
                circ_ss_idx[parser.contig][parser.start][parser.strand]['start'] = 1
                circ_ss_idx[parser.contig][parser.end][parser.strand]['end'] = 1
    elif circ_path.suffix == ".bed":
        LOGGER.info('Loading additional circRNA bed ..')
        n_skip = 0
        with open(circ_path, 'r') as f:
            for line in f:
                content = line.rstrip().split('\t')
                contig = content[0]
                try:
                    start = int(content[1])
                    end = int(content[2])
                except ValueError:
                    n_skip += 1
                    continue
                strand = content[3]
                circ_ss_idx[contig][start][strand]['start'] = 1
                circ_ss_idx[contig][end][strand]['end'] = 1
        LOGGER.warn('Skipping {} lines in bed file'.format(n_skip))
    else:
        sys.exit('{} is not a valid bed/gtf file'.format(str(circ_path)))

    return circ_ss_idx


def get_blocks(hit):
    """
    Get blocks of aligned segments
    :param hit:
    :return:
    """
    r_start = hit.r_st
    r_end = hit.r_st
    r_block = []
    for length, operation in hit.cigar:
        if operation == 0:
            r_end += length
        elif operation == 1:
            pass
        elif operation == 2:
            r_end += length
        elif operation == 3:
            r_block.append([r_start, r_end, r_end - r_start + 1])
            r_start = r_end + length
            r_end = r_start
        elif operation == 4:
            pass
    if r_end > r_start:
        r_block.append([r_start, r_end, r_end - r_start + 1])
    return r_block


def get_exons(hit):
    r_start, r_end = hit.r_st, hit.r_st
    q_start, q_end = hit.q_st, hit.q_st
    r_block = []
    for length, operation in hit.cigar:
        if operation == 0:
            r_end += length
            q_end += length
        elif operation == 1:
            q_end += length
        elif operation == 2:
            r_end += length
        elif operation == 3:
            r_block.append([r_start, r_end, q_start, q_end])
            r_start = r_end + length
            r_end = r_start
            q_start = q_end
        elif operation == 4:
            pass

    if r_end > r_start:
        r_block.append([r_start, r_end, q_start, q_end])
    else:
        pass

    return r_block


def get_parital_blocks(hit, junc):
    exons = get_exons(hit)
    blocks = []
    for r_st, r_en, q_st, q_en in exons:
        if abs(q_st - junc) <= 10:
            blocks.append([r_st, r_en, '*-'])
        elif abs(q_en - junc) <= 10:
            blocks.append([r_st, r_en, '-*'])
        else:
            blocks.append([r_st, r_en, r_en - r_st + 1])
    return blocks


def merge_blocks(blocks):
    from operator import itemgetter
    tmp = sorted(blocks, key=itemgetter(0, 1))
    merged = []
    last_st, last_en = tmp[0][0], tmp[0][1]
    for st, en, length in tmp[1:]:
        if st <= last_en:
            last_en = max(en, last_en)
            last_st = min(st, last_st)
        else:
            merged.append([last_st, last_en, last_en - last_st + 1])
            last_st, last_en = st, en
    merged.append([last_st, last_en, last_en - last_st + 1])
    return merged


def merge_exons(tail_exons, head_exons):
    if head_exons[0][0] < tail_exons[-1][1]:
        return merge_blocks(tail_exons + head_exons)
    else:
        head_exons[0] = [head_exons[0][0], head_exons[0][1], '*-']
        tail_exons[-1] = [tail_exons[-1][0], tail_exons[-1][1], '-*']
        return tail_exons + head_exons


def merge_clip_exon(exons, clip_info):
    clip_st, clip_en = clip_info[0], clip_info[1]
    exon_st, exon_en = exons[0][0], exons[-1][1]

    if clip_st and clip_en:
        if clip_en < exon_st:
            exons = [[clip_st, clip_en, clip_en - clip_st + 1], ] + exons
        elif exon_en < clip_st:
            exons = exons + [[clip_st, clip_en, clip_en - clip_st + 1], ]
        elif clip_st < exon_st < clip_en:
            exons[0] = [clip_st, exons[0][1], exons[0][1] - clip_st + 1]
        elif clip_st < exon_en < clip_en:
            exons[-1] = [exons[-1][0], clip_en, clip_en - exons[-1][0] + 1]
        else:
            pass

    return exons


def remove_long_insert(hit):
    r_st, q_st = hit.r_st, hit.q_st
    last_r_st, last_q_st = r_st, q_st
    last_cigar = []
    sub_hits = []
    for length, operation in hit.cigar:
        if operation == 0:
            r_st += length
            q_st += length
        elif operation == 1:
            q_st += length
            if length > 20:
                sub_hits.append(SubHit(hit, last_r_st, last_q_st, last_cigar))
                last_cigar = []
                last_r_st, last_q_st = r_st, q_st
                continue
        elif operation in [2, 3]:
            r_st += length
        elif operation in [4, 5]:
            if q_st == hit.q_st:
                q_st += length
        else:
            pass
        last_cigar.append((length, operation))
    if last_cigar:
        sub_hits.append(SubHit(hit, last_r_st, last_q_st, last_cigar))
    primary_hit = sorted(sub_hits, key=lambda x: x.mlen, reverse=True)[0]
    primary_hit.is_primary = 1

    return primary_hit


def get_primary_alignment(hits):
    if not hits:
        return None

    for hit in hits:
        if hit.is_primary:
            return remove_long_insert(hit)

    return None


def find_annotated_signal(contig, start, end, clip_base, search_length=10, shift_threshold=3):
    tmp_annotated_signal = {}

    ds_free = 0
    for i in range(100):
        if end + i > env.CONTIG_LEN[contig]:
            break
        if env.GENOME.seq(contig, start, start + i) == env.GENOME.seq(contig, end, end + i):
            ds_free = i
        else:
            break

    us_free = 0
    for j in range(100):
        if start - j < 0:
            break
        if env.GENOME.seq(contig, start - j, start) == env.GENOME.seq(contig, end - j, end):
            us_free = j
        else:
            break

    if start - search_length - us_free - 2 < 0 or end + search_length + ds_free + 2 > env.CONTIG_LEN[contig]:
        return None, us_free, ds_free, tmp_annotated_signal

    # First: find annotated splice signal
    if env.SS_INDEX is not None and contig in env.SS_INDEX:
        anno_ss = []
        for strand in ['+', '-']:
            # Upstream
            tmp_us_sites = []
            for us_shift in range(-search_length, search_length):
                us_pos = start + us_shift + 1
                if us_pos not in env.SS_INDEX[contig]:
                    continue
                if strand not in env.SS_INDEX[contig][us_pos]:
                    continue
                if 'start' not in env.SS_INDEX[contig][us_pos][strand]:
                    continue
                tmp_us_sites.append(us_shift)

            for us_shift in range(-search_length, search_length):
                us_pos = start + us_shift
                if us_pos not in env.SS_INDEX[contig]:
                    continue
                if strand not in env.SS_INDEX[contig][us_pos]:
                    continue
                if 'end' in env.SS_INDEX[contig][us_pos][strand]:
                    tmp_us_sites.append(us_shift)

            # Downstream
            tmp_ds_sites = []
            for ds_shift in range(-search_length, search_length):
                ds_pos = end + ds_shift + 1
                if ds_pos not in env.SS_INDEX[contig]:
                    continue
                if strand not in env.SS_INDEX[contig][ds_pos]:
                    continue
                if 'start' not in env.SS_INDEX[contig][ds_pos][strand]:
                    continue
                tmp_ds_sites.append(ds_shift)

            for ds_shift in range(-search_length, search_length):
                ds_pos = end + ds_shift
                if ds_pos not in env.SS_INDEX[contig]:
                    continue
                if strand not in env.SS_INDEX[contig][ds_pos]:
                    continue
                if 'end' not in env.SS_INDEX[contig][ds_pos][strand]:
                    continue
                tmp_ds_sites.append(ds_shift)

            tmp_annotated_signal[strand] = (tmp_us_sites, tmp_ds_sites)

            if len(tmp_us_sites) == 0 or len(tmp_ds_sites) == 0:
                continue

            for i in tmp_us_sites:
                for j in tmp_ds_sites:
                    if abs(i - j) > shift_threshold + clip_base:
                        continue
                    us_ss = env.GENOME.seq(contig, start + i - 2, start + i)
                    ds_ss = env.GENOME.seq(contig, end + j, end + j + 2)
                    if strand == '-':
                        us_ss, ds_ss = revcomp(ds_ss), revcomp(us_ss)
                    ss_id = '{}-{}|{}-{}'.format(us_ss, ds_ss, i, j)
                    ss_weight = SPLICE_SIGNAL[(ds_ss, us_ss)] if (ds_ss, us_ss) in SPLICE_SIGNAL else 3

                    anno_ss.append((
                        ss_id, strand, i, j, ss_weight, *get_ss_altered_length(i, j, us_free, ds_free, clip_base)
                    ))

        if len(anno_ss) > 0:
            return sort_ss(anno_ss, us_free, ds_free, clip_base), us_free, ds_free, tmp_annotated_signal

    return None, us_free, ds_free, tmp_annotated_signal


def find_denovo_signal(contig, start, end, host_strand, tmp_signal, us_free, ds_free, clip_base,
                       search_length=10, shift_threshold=3, is_canonical=False):
    # Second: find GT-AG splice signal
    us_search_length = search_length + us_free
    ds_search_length = search_length + ds_free
    us_seq = env.GENOME.seq(contig, start - us_search_length - 2, start + ds_search_length)
    ds_seq = env.GENOME.seq(contig, end - us_search_length, end + ds_search_length + 2)

    if us_seq is None or len(us_seq) < ds_search_length - us_search_length + 2:
        return None
    if ds_seq is None or len(ds_seq) < ds_search_length - us_search_length + 2:
        return None

    # Same strand
    prior_ss = []
    if host_strand:
        prior_strand = set(list(host_strand))
        for strand in prior_strand:
            for (tmp_ds_ss, tmp_us_ss), ss_weight in SPLICE_SIGNAL.items():
                # Only search canonical GT-AG
                if is_canonical and ss_weight != 0:
                    continue
                if strand == '-':
                    ds_ss, us_ss = revcomp(tmp_us_ss), revcomp(tmp_ds_ss)
                else:
                    ds_ss, us_ss = tmp_ds_ss, tmp_us_ss

                # Find upstream signal
                tmp_us_start = 0
                tmp_us_sites = []
                while 1:
                    tmp_us = us_seq.find(us_ss, tmp_us_start + 1)
                    if tmp_us == -1:
                        break
                    tmp_us_sites.append(tmp_us - us_search_length)
                    tmp_us_start = tmp_us

                # Find downstream signal
                tmp_ds_start = 0
                tmp_ds_sites = []
                while 1:
                    tmp_ds = ds_seq.find(ds_ss, tmp_ds_start + 1)
                    if tmp_ds == -1:
                        break
                    tmp_ds_sites.append(tmp_ds - us_search_length)
                    tmp_ds_start = tmp_ds

                if strand in tmp_signal:
                    tmp_us_signal, tmp_ds_signal = tmp_signal[strand]
                    tmp_us_sites = sorted(list(set(tmp_us_sites + tmp_us_signal)))
                    tmp_ds_sites = sorted(list(set(tmp_ds_sites + tmp_ds_signal)))

                # Filter paired splice signal in concordance position
                if len(tmp_us_sites) == 0 or len(tmp_ds_sites) == 0:
                    continue

                for i in tmp_us_sites:
                    for j in tmp_ds_sites:
                        if abs(i - j) > clip_base + shift_threshold:
                            continue
                        ss_id = '{}-{}*|{}-{}'.format(tmp_us_ss, tmp_ds_ss, i, j)
                        prior_ss.append((
                            ss_id, strand, i, j, ss_weight,
                            *get_ss_altered_length(i, j, us_free, ds_free, clip_base)
                        ))

    if len(prior_ss) > 0:
        return sort_ss(prior_ss, us_free, ds_free, clip_base)

    # Anti-sense
    other_ss = []
    other_strand = {'+', '-'} - set(list(host_strand)) if host_strand else {'+', '-'}
    if other_strand:
        for strand in other_strand:
            for (tmp_ds_ss, tmp_us_ss), ss_weight in SPLICE_SIGNAL.items():
                if is_canonical and ss_weight != 0:
                    continue
                if strand == '-':
                    ds_ss, us_ss = revcomp(tmp_us_ss), revcomp(tmp_ds_ss)
                else:
                    ds_ss, us_ss = tmp_ds_ss, tmp_us_ss

                # Find upstream signal
                tmp_us_start = 0
                tmp_us_sites = []
                while 1:
                    tmp_us = us_seq.find(us_ss, tmp_us_start + 1)
                    if tmp_us == -1:
                        break
                    tmp_us_sites.append(tmp_us - us_search_length)
                    tmp_us_start = tmp_us

                # Find downstream signal
                tmp_ds_start = 0
                tmp_ds_sites = []
                while 1:
                    tmp_ds = ds_seq.find(ds_ss, tmp_ds_start + 1)
                    if tmp_ds == -1:
                        break
                    tmp_ds_sites.append(tmp_ds - us_search_length)
                    tmp_ds_start = tmp_ds

                if strand in tmp_signal:
                    tmp_us_signal, tmp_ds_signal = tmp_signal[strand]
                    tmp_us_sites = sorted(list(set(tmp_us_sites + tmp_us_signal)))
                    tmp_ds_sites = sorted(list(set(tmp_ds_sites + tmp_ds_signal)))

                # Filter paired splice signal in concordance position
                if len(tmp_us_sites) == 0 or len(tmp_ds_sites) == 0:
                    continue

                for i in tmp_us_sites:
                    for j in tmp_ds_sites:
                        if abs(i - j) > clip_base + shift_threshold:
                            continue
                        ss_id = '{}-{}*|{}-{}'.format(tmp_us_ss, tmp_ds_ss, i, j)
                        other_ss.append((
                            ss_id, strand, i, j, ss_weight,
                            *get_ss_altered_length(i, j, us_free, ds_free, clip_base)
                        ))

    if len(other_ss) > 0:
        return sort_ss(other_ss, us_free, ds_free, clip_base)

    return None


def get_ss_altered_length(i, j, us_free, ds_free, clip_base):
    clip_altered = min(abs(j - i - clip_base), abs(j - i + clip_base))
    us_altered = min(abs(i + us_free), abs(i - ds_free))
    ds_altered = min(abs(j + us_free), abs(j - ds_free))
    return abs(i-j), clip_altered, us_altered + ds_altered


def sort_ss(sites, us, ds, clip_base):
    # Splice site: site_id, strand, us_shift, ds_shift, site_weight, altered_len, clip_altered, altered_total
    from operator import itemgetter
    get_ss = itemgetter(0, 1, 2, 3)

    tmp_sites = set(sites)

    # Clipped sites
    clipped_sites = [i for i in tmp_sites if (-clip_base <= i[2] - i[3] <= clip_base)]
    if len(clipped_sites) > 0:
        return get_ss(sorted(clipped_sites, key=itemgetter(6, 5, 4, 7))[0])
    tmp_sites = set(sites) - set(clipped_sites)

    # Confidential splice sites
    confident_sites = [i for i in tmp_sites if -us <= i[2] <= ds and -us <= i[3] <= ds]
    if len(confident_sites) > 0:
        return get_ss(sorted(confident_sites, key=itemgetter(5, 4, 6, 7))[0])
    tmp_sites = tmp_sites - set(confident_sites)

    # Ambiguous splice sites
    ambiguous_sites = [i for i in tmp_sites if -clip_base <= i[2] <= 0 <= i[3] <= clip_base]
    if len(ambiguous_sites) > 0:
        return get_ss(sorted(ambiguous_sites, key=itemgetter(4, 5, 6, 7))[0])
    tmp_sites = tmp_sites - set(ambiguous_sites)

    # Other sites
    if len(tmp_sites) > 0:
        return get_ss(sorted(tmp_sites, key=itemgetter(4, 5, 6, 7))[0])
    return None


def find_host_gene(ctg, start, end):
    if ctg not in env.GTF_INDEX:
        return None
    start_div, end_div = start // 500, end // 500

    host_gene = {}
    for x in range(start_div, end_div + 1):
        if x not in env.GTF_INDEX[ctg]:
            continue
        for element in env.GTF_INDEX[ctg][x]:
            # start site
            if element.end < start or element.start > end:
                continue
            if element.start - 500 <= start <= element.end + 500 or element.start - 500 <= end <= element.end + 500:
                host_gene.setdefault(element.strand, []).append(element)

    if host_gene:
        return host_gene
    else:
        return None


def find_retained_introns(ctg, start, end):
    if ctg not in env.INTRON_INDEX:
        return None
    start_div, end_div = start // 500, end // 500

    host_gene = {}
    for x in range(start_div, end_div + 1):
        if x not in env.INTRON_INDEX[ctg]:
            continue
        for st, en, strand in env.INTRON_INDEX[ctg][x]:
            if st - 25 <= start and end <= en + 25:
                host_gene.setdefault(strand, []).append((st, en, strand))

    if host_gene:
        return host_gene
    else:
        return None


def find_overlap_exons(ctg, start, end):
    if ctg not in env.GTF_INDEX:
        return None
    start_div, end_div = start // 500, end // 500

    host_gene = {}
    for x in range(start_div, end_div + 1):
        if x not in env.GTF_INDEX[ctg]:
            continue
        for element in env.GTF_INDEX[ctg][x]:
            if element.type != 'exon':
                continue
            if element.end - 25 < start or end < element.start + 25:
                continue
            host_gene.setdefault(element.strand, []).append((element.start, element.end, element.strand))

    if host_gene:
        return host_gene
    else:
        return None


def convert_cigar_string(x):
    return [(int(l), OPERATION[op]) for l, op in re.findall(r'(\d+)([MIDNSHP=X])', x)]


def find_alignment_pos(alignment, pos):
    r_st, r_en = alignment.ref_begin, alignment.ref_begin
    q_st, q_en = alignment.query_begin, alignment.query_begin
    for l, op in convert_cigar_string(alignment.cigar_string):
        if op == 0:
            r_en += l
            q_en += l
        elif op == 1:
            q_en += l
        elif op in [2, ]:
            r_en += l
        elif op in [4, 5]:
            pass
        if r_st <= pos <= r_en:
            return q_st + pos - r_st
        r_st = r_en
        q_st = q_en
    return None



ALIGNER = None
CONTIG_LEN = None
GENOME = None
GTF_INDEX = None
INTRON_INDEX = None
SS_INDEX = None


def initializer(aligner, contig_len, genome, gtf_index, intron_index, ss_index):
    global ALIGNER
    global CONTIG_LEN
    global GENOME
    global GTF_INDEX
    global INTRON_INDEX
    global SS_INDEX
    ALIGNER = aligner
    CONTIG_LEN = contig_len
    GENOME = genome
    GTF_INDEX = gtf_index
    INTRON_INDEX = intron_index
    SS_INDEX = ss_index





#! /usr/bin/env python
# -*- encoding:utf-8 -*-
import os
import sys
sys.path.append('CIRI_long')
import codecs
from pathlib import Path
from subprocess import getstatusoutput
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install

from CIRI_long.version import __version__


class CustomInstall(install):
    def run(self):
        # Build library
        lib_dir = Path(os.path.abspath(__file__)).parent / 'libs' / 'striped_smith_waterman'
        status, ret = getstatusoutput('cd {} && make libssw.so'.format(lib_dir))
        if status != 0:
            sys.exit('Build Error: {}'.format(ret))
        else:
            print(ret)
        # Install
        install.run(self)


def read(infile):
    return codecs.open(os.path.join(os.path.dirname(__file__), infile)).read()

setup(
    name='CIRI-long',
    version=__version__,
    url='https://github.com/bioinfo-biols/CIRI-long',
    description='circular RNA identification from Nanopore',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Jinyang Zhang',
    author_email='zhangjinyang@biols.ac.cn',
    maintainer='Jinyang Zhang',
    maintainer_email='zhangjinyang@biols.ac.cn',
    license='MIT',
    keywords='circRNA',
    packages=find_packages(exclude=['doc', 'tests']),
    entry_points={
        'console_scripts': [
            'CIRI-long=CIRI_long.main:main',
        ]
    },
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'argparse>=1.2.1', 'Cython==0.29.13', 'mappy==2.17', 'numpy==1.17.0', 'pandas==0.25.0',
        'pysam==0.15.3', 'python-Levenshtein==0.12.1', 'scikit-learn==0.21.3', 'scipy==1.3.1',
        'biopython==1.76', 'pyspoa==0.0.5', 'bwapy==0.1.4',
    ],
    test_suite="nose.collector",
    tests_require=['nose==1.3.7'],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
    ],
    cmdclass={
        'build_ext': build_ext,
        'install': CustomInstall,
    }
)



import os
import sys
import re
import logging
from multiprocessing import Pool
from collections import defaultdict

import pysam
from CIRI_long import env
from CIRI_long.align import *
from CIRI_long.logger import ProgressBar
from CIRI_long.utils import grouper, revcomp

LOGGER = logging.getLogger('CIRI-long')


def search_splice_signal(contig, start, end, clip_base, search_length=10, shift_threshold=3):
    # Find free sliding region
    # start | real_start <-> end | real_end
    ds_free = 0
    for i in range(100):
        if end + i > env.CONTIG_LEN[contig]:
            break
        if env.GENOME.seq(contig, start, start + i) == env.GENOME.seq(contig, end, end + i):
            ds_free = i
        else:
            break

    us_free = 0
    for j in range(100):
        if start - j < 0:
            break
        if env.GENOME.seq(contig, start - j, start) == env.GENOME.seq(contig, end - j, end):
            us_free = j
        else:
            break

    if start - search_length - us_free - 2 < 0 or end + search_length + ds_free + 2 > env.CONTIG_LEN[contig]:
        return None, us_free, ds_free
    # Splice site: site_id, strand, us_shift, ds_shift, site_weight, altered_len, altered_total
    # First: Find flanking junction from annotation gtf
    if env.SS_INDEX is not None:
        anno_ss = []
        for strand in ['+', '-']:
            tmp_us_sites = []
            for us_shift in range(-search_length, search_length):
                us_pos = start + us_shift
                if contig in env.SS_INDEX and us_pos in env.SS_INDEX[contig] and strand in env.SS_INDEX[contig][us_pos]:
                    tmp_us_sites.append(us_shift - 1)

            tmp_ds_sites = []
            for ds_shift in range(-search_length, search_length):
                ds_pos = end + ds_shift
                if contig in env.SS_INDEX and ds_pos in env.SS_INDEX[contig] and strand in env.SS_INDEX[contig][ds_pos]:
                    tmp_ds_sites.append(ds_shift)

            if len(tmp_us_sites) == 0 or len(tmp_ds_sites) == 0:
                continue

            for i in tmp_us_sites:
                for j in tmp_ds_sites:
                    if abs(i - j) > shift_threshold + clip_base:
                        continue
                    us_ss = env.GENOME.seq(contig, start + i - 2, start + i)
                    ds_ss = env.GENOME.seq(contig, end + j, end + j + 2)
                    if strand == '-':
                        us_ss, ds_ss = revcomp(ds_ss), revcomp(us_ss)
                    ss_id = '{}-{}|{}-{}'.format(us_ss, ds_ss, i, j)
                    ss_weight = SPLICE_SIGNAL[(ds_ss, us_ss)] if (ds_ss, us_ss) in SPLICE_SIGNAL else 3

                    anno_ss.append((
                        ss_id, strand, i, j, ss_weight, *get_ss_altered_length(i, j, us_free, ds_free, clip_base)
                    ))

        if len(anno_ss) > 0:
            return sort_ss(anno_ss, us_free, ds_free, clip_base), us_free, ds_free

    # Second: Find Denovo BSJ using pre-defined splice signal
    us_search_length = search_length + us_free
    ds_search_length = search_length + ds_free
    us_seq = env.GENOME.seq(contig, start - us_search_length - 2, start + ds_search_length)
    ds_seq = env.GENOME.seq(contig, end - us_search_length, end + ds_search_length + 2)

    if us_seq is None or len(us_seq) < ds_search_length - us_search_length + 2:
        return None, us_free, ds_free
    if ds_seq is None or len(ds_seq) < ds_search_length - us_search_length + 2:
        return None, us_free, ds_free

    putative_ss = []
    for strand in ['+', '-']:
        for (tmp_ds_ss, tmp_us_ss), ss_weight in SPLICE_SIGNAL.items():
            if strand == '-':
                ds_ss, us_ss = revcomp(tmp_us_ss), revcomp(tmp_ds_ss)
            else:
                ds_ss, us_ss = tmp_ds_ss, tmp_us_ss

            # Find upstream signal
            tmp_us_start = 0
            tmp_us_sites = []
            while 1:
                tmp_us = us_seq.find(us_ss, tmp_us_start + 1)
                if tmp_us == -1:
                    break
                tmp_us_sites.append(tmp_us)
                tmp_us_start = tmp_us

            # Find downstream signal
            tmp_ds_start = 0
            tmp_ds_sites = []
            while 1:
                tmp_ds = ds_seq.find(ds_ss, tmp_ds_start + 1)
                if tmp_ds == -1:
                    break
                tmp_ds_sites.append(tmp_ds)
                tmp_ds_start = tmp_ds

            # Filter paired splice signal in concordance position
            if len(tmp_us_sites) == 0 or len(tmp_ds_sites) == 0:
                continue

            for i in tmp_us_sites:
                for j in tmp_ds_sites:
                    if abs(i - j) > clip_base + shift_threshold:
                        continue
                    us_shift = i - us_search_length
                    ds_shift = j - us_search_length
                    ss_id = '{}-{}*|{}-{}'.format(tmp_us_ss, tmp_ds_ss, us_shift, ds_shift)
                    putative_ss.append((
                        ss_id, strand, us_shift, ds_shift, ss_weight,
                        *get_ss_altered_length(us_shift, ds_shift, us_free, ds_free, clip_base)
                    ))

    if len(putative_ss) > 0:
        return sort_ss(putative_ss, us_free, ds_free, clip_base), us_free, ds_free

    return None, us_free, ds_free


def find_bsj(ccs):
    """
    Find junction using aligner
    """
    init_hit = get_primary_alignment(env.ALIGNER.map(ccs * 2))
    if init_hit is None:
        return None, None

    circ_junc = init_hit.q_st % len(ccs)
    circ = ccs[circ_junc:] + ccs[:circ_junc]

    last_junc = 0
    last_m = 0

    itered_junc = {}
    while True:
        circ_hit = get_primary_alignment(env.ALIGNER.map(circ))
        if circ_hit is None or circ_hit.mlen <= last_m:
            circ_junc = last_junc
            break
        last_m = circ_hit.mlen
        last_junc = circ_junc

        st_clip, en_clip = circ_hit.q_st, len(circ) - circ_hit.q_en
        if st_clip == 0 and en_clip == 0:
            break

        if st_clip >= en_clip:
            circ_junc = (circ_junc + st_clip) % len(circ)
        else:
            circ_junc = (circ_junc + circ_hit.q_en) % len(circ)

        if circ_junc in itered_junc:
            circ_junc = last_junc
            break

        circ = ccs[circ_junc:] + ccs[:circ_junc]
        itered_junc[circ_junc] = 1

    circ = ccs[circ_junc:] + ccs[:circ_junc]
    return circ, circ_junc


def align_clip_segments(circ, hit):
    """
    Align clip bases
    """
    from libs.striped_smith_waterman.ssw_wrap import Aligner
    from collections import Counter
    st_clip, en_clip = hit.q_st, len(circ) - hit.q_en
    clip_r_st, clip_r_en, clipped_circ = None, None, None

    if st_clip + en_clip >= 20:
        clip_seq = circ[hit.q_en:] + circ[:hit.q_st]
        if len(clip_seq) > 0.6 * len(circ):
            return None, None, None, None

        tmp_start = max(hit.r_st - 200000, 0)
        tmp_end = min(hit.r_en + 200000, env.CONTIG_LEN[hit.ctg])

        tmp_seq = env.GENOME.seq(hit.ctg, tmp_start, tmp_end)
        if Counter(tmp_seq)['N'] >= 0.3 * (tmp_end - tmp_start):
            return None, None, None, None

        if hit.strand > 0:
            ssw = Aligner(tmp_seq, match=1, mismatch=1, gap_open=1, gap_extend=1)
            align_res = ssw.align(clip_seq)
            clip_r_st, clip_r_en = tmp_start + align_res.ref_begin, tmp_start + align_res.ref_end
            if clip_r_st < hit.r_st:
                clipped_circ = clip_seq[align_res.query_begin:] + \
                               circ[hit.q_st:hit.q_en] + \
                               clip_seq[:align_res.query_begin]
            else:
                clipped_circ = circ[hit.q_st:] + circ[:hit.q_st]
        else:
            ssw = Aligner(revcomp(tmp_seq), match=1, mismatch=1, gap_open=1, gap_extend=1)
            align_res = ssw.align(clip_seq)
            clip_r_st, clip_r_en = tmp_end - align_res.ref_end, tmp_end - align_res.ref_begin
            if clip_r_en > hit.r_en:
                clipped_circ = clip_seq[align_res.query_begin:] + \
                               circ[hit.q_st:hit.q_en] + \
                               clip_seq[:align_res.query_begin]
            else:
                clipped_circ = circ[hit.q_st:] + circ[:hit.q_st]

        clip_base = hit.q_st + len(circ) - hit.q_en - (align_res.query_end - align_res.query_begin) + 1
        circ_start = min(hit.r_st, clip_r_st) - 1
        circ_end = max(hit.r_en, clip_r_en)
    else:
        clipped_circ = circ[hit.q_st:] + circ[:hit.q_st]
        clip_base = st_clip + en_clip
        circ_start = hit.r_st - 1
        circ_end = hit.r_en

    return clipped_circ, circ_start, circ_end, (clip_r_st, clip_r_en, clip_base)


def scan_ccs_chunk(chunk, is_canonical):
    reads_cnt = defaultdict(int)
    ret = []

    short_reads = []
    for read_id, segments, ccs, raw in chunk:
        # Filter 1 - Remove linear mapped reads
        raw_hit = get_primary_alignment(env.ALIGNER.map(raw))
        if raw_hit and raw_hit.mlen > max(len(raw) * 0.8, len(raw) - 200):
            continue
        if raw_hit and raw_hit.mlen > 1.5 * len(ccs):
            continue

        raw_st = raw_hit.q_st if raw_hit else None
        raw_en = raw_hit.q_en if raw_hit else None
        reads_cnt['raw_unmapped'] += 1

        # Filter 2 - Remove other mapped region that intersect with ccs
        seg_st = int(segments.split(';')[0].split('-')[0])
        seg_en = int(segments.split(';')[-1].split('-')[1])
        if raw_hit and (raw_en < seg_st or raw_st > seg_en):
            continue

        ccs_hit = get_primary_alignment(env.ALIGNER.map(ccs * 2))
        if ccs_hit is None and len(ccs) < 150:
            short_reads.append((read_id, segments, ccs, raw))
        if ccs_hit is None or seg_en - seg_st < ccs_hit.q_en - ccs_hit.q_st:
            continue

        reads_cnt['ccs_mapped'] += 1

        # Find back-spliced junction site
        circ, junc = find_bsj(ccs)

        # Candidate alignment situation, more than 85%
        circ_hit = get_primary_alignment(env.ALIGNER.map(circ))
        if circ_hit is None or circ_hit.mlen < 0.75 * len(circ):
            continue

        clipped_circ, circ_start, circ_end, clip_info = align_clip_segments(circ, circ_hit)
        if circ_start is None or circ_end is None:
            continue

        clip_base = clip_info[2]
        if clip_base > 0.15 * len(ccs) or clip_base > 20:
            continue

        reads_cnt['bsj'] += 1

        # Retrive circRNA positions, convert minimap2 position to real position
        host_strand = find_host_gene(circ_hit.ctg, circ_start, circ_end)
        ss_site, us_free, ds_free, tmp_signal = find_annotated_signal(circ_hit.ctg, circ_start, circ_end, clip_base, clip_base + 10)
        if ss_site is None:
            ss_site = find_denovo_signal(circ_hit.ctg, circ_start, circ_end, host_strand, tmp_signal,
                                         us_free, ds_free, clip_base, clip_base + 10, 3, True)

        if ss_site is None:
            ss_id = 'NA'
            strand = 'NA'
            correction_shift = 0
        else:
            reads_cnt['signal'] += 1
            ss_id, strand, us_shift, ds_shift = ss_site
            circ_start += us_shift
            circ_end += ds_shift
            correction_shift = min(max(us_shift, us_free), ds_free)

        circ_id = '{}:{}-{}'.format(circ_hit.ctg, circ_start + 1, circ_end)

        # Get Cirexons
        cir_exons = get_blocks(circ_hit)
        cir_exons = merge_clip_exon(cir_exons, clip_info)

        cir_exons[0][0] = circ_start
        cir_exons[-1][1] = circ_end

        cir_exon_tag = []
        for cir_exon_start, cir_exon_end, cir_exon_length in cir_exons:
            cir_exon_tag.append('{}-{}|{}'.format(cir_exon_start + 1, cir_exon_end, cir_exon_length))

        # BSJ correction for 5' prime region
        circ_seq = clipped_circ if circ_hit.strand > 0 else revcomp(clipped_circ)
        circ_seq = circ_seq[correction_shift:] + circ_seq[:correction_shift]

        ret.append((
            read_id, circ_id, strand, ','.join(cir_exon_tag), ss_id,
            '{}|{}-{}'.format(junc, clip_base, len(circ)), segments, circ_seq
        ))

    return reads_cnt, short_reads, ret


def scan_ccs_reads(ccs_seq, ref_fasta, ss_index, gtf_index, intron_index, is_canonical, out_dir, prefix, threads):
    import mappy as mp

    faidx = Faidx(ref_fasta)
    contig_len = faidx.contig_len
    faidx.close()

    # First scanning using minimap2
    minimap_aligner = mp.Aligner(ref_fasta, n_threads=threads, preset='splice')

    chunk_size = 250
    jobs = []
    pool = Pool(threads, env.initializer,
                (minimap_aligner, contig_len, minimap_aligner, gtf_index, intron_index, ss_index))

    for reads in grouper(list(ccs_seq), chunk_size):
        chunk = [[i, ] + ccs_seq[i] for i in reads if i is not None]
        jobs.append(pool.apply_async(scan_ccs_chunk, (chunk, is_canonical)))
    pool.close()

    prog = ProgressBar()
    finished_cnt = 0

    reads_count = defaultdict(int)
    short_reads = []
    with open('{}/{}.cand_circ.fa'.format(out_dir, prefix), 'w') as out:
        for job in jobs:
            finished_cnt += 1

            tmp_cnt, tmp_short, ret = job.get()
            for key, value in tmp_cnt.items():
                reads_count[key] += value

            short_reads += tmp_short

            for read_id, circ_id, strand, cir_exon_tag, ss_id, clip_info, segments, circ_seq in ret:
                out.write('>{}\t{}\t{}\t{}\t{}\t{}\t{}\n{}\n'.format(
                    read_id, circ_id, strand, cir_exon_tag, ss_id, clip_info, segments, circ_seq
                ))
            prog.update(100 * finished_cnt / len(jobs))

    pool.join()
    prog.update(100)

    return reads_count, short_reads


def recover_ccs_chunk(chunk, is_canonical):
    reads_cnt = defaultdict(int)
    ret = []

    for read_id, segments, ccs, raw in chunk:
        # Remove other mapped region that intersect with ccs
        seg_st = int(segments.split(';')[0].split('-')[0])
        seg_en = int(segments.split(';')[-1].split('-')[1])

        ccs_hit = get_primary_alignment(env.ALIGNER.map(ccs * 2))
        if ccs_hit is None or seg_en - seg_st < ccs_hit.q_en - ccs_hit.q_st:
            continue

        reads_cnt['ccs_mapped'] += 1

        # Find back-spliced junction site
        circ, junc = find_bsj(ccs)

        # Candidate alignment situation, more than 85%
        circ_hit = get_primary_alignment(env.ALIGNER.map(circ))
        if circ_hit is None:
            continue

        clipped_circ, circ_start, circ_end, clip_info = align_clip_segments(circ, circ_hit)
        if circ_start is None or circ_end is None:
            continue

        clip_base = clip_info[2]
        if clip_base > 0.15 * len(ccs) or clip_base > 20:
            continue

        reads_cnt['bsj'] += 1

        # Retrive circRNA positions, convert minimap2 position to real position
        host_strand = find_host_gene(circ_hit.ctg, circ_start, circ_end)
        ss_site, us_free, ds_free, tmp_signal = find_annotated_signal(circ_hit.ctg, circ_start, circ_end, clip_base, clip_base + 10)
        if ss_site is None:
            ss_site = find_denovo_signal(circ_hit.ctg, circ_start, circ_end, host_strand, tmp_signal,
                                         us_free, ds_free, clip_base, clip_base + 10, 3, True)

        if ss_site is None:
            ss_id = 'NA'
            strand = 'NA'
            correction_shift = 0
        else:
            reads_cnt['signal'] += 1
            ss_id, strand, us_shift, ds_shift = ss_site
            circ_start += us_shift
            circ_end += ds_shift
            correction_shift = min(max(us_shift, us_free), ds_free)

        circ_id = '{}:{}-{}'.format(circ_hit.ctg, circ_start + 1, circ_end)

        # Get Cirexons
        cir_exons = get_blocks(circ_hit)
        cir_exons = merge_clip_exon(cir_exons, clip_info)

        cir_exons[0][0] = circ_start
        cir_exons[-1][1] = circ_end

        cir_exon_tag = []
        for cir_exon_start, cir_exon_end, cir_exon_length in cir_exons:
            cir_exon_tag.append('{}-{}|{}'.format(cir_exon_start + 1, cir_exon_end, cir_exon_length))

        # BSJ correction for 5' prime region
        circ_seq = clipped_circ if circ_hit.strand > 0 else revcomp(clipped_circ)
        circ_seq = circ_seq[correction_shift:] + circ_seq[:correction_shift]

        ret.append((
            read_id, circ_id, strand, ','.join(cir_exon_tag), ss_id,
            '{}|{}-{}'.format(junc, clip_base, len(circ)), segments, circ_seq
        ))

    return reads_cnt, ret


def recover_ccs_reads(short_reads, ref_fasta, ss_index, gtf_index, intron_index, is_canonical, out_dir, prefix, threads):
    from bwapy import BwaAligner

    # Second scanning of short reads
    genome = Fasta(ref_fasta)

    options = '-x ont2d -T 19'
    bwa_aligner = Aligner(BwaAligner(ref_fasta, options=options))

    chunk_size = 250
    jobs = []
    pool = Pool(threads, env.initializer, (bwa_aligner, genome.contig_len, genome, gtf_index, intron_index, ss_index))
    for reads in grouper(short_reads, chunk_size):
        chunk = [i for i in reads if i is not None]
        jobs.append(pool.apply_async(recover_ccs_chunk, (chunk, is_canonical)))
    pool.close()

    prog = ProgressBar()
    prog.update(0)
    finished_cnt = 0

    reads_count = defaultdict(int)
    with open('{}/{}.cand_circ.fa'.format(out_dir, prefix), 'a') as out:
        for job in jobs:
            finished_cnt += 1

            tmp_cnt, ret = job.get()
            for key, value in tmp_cnt.items():
                reads_count[key] += value

            for read_id, circ_id, strand, cir_exon_tag, ss_id, clip_info, segments, circ_seq in ret:
                out.write('>{}\t{}\t{}\t{}\t{}\t{}\t{}\n{}\n'.format(
                    read_id, circ_id, strand, cir_exon_tag, ss_id, clip_info, segments, circ_seq
                ))
            prog.update(100 * finished_cnt / len(jobs))

    pool.join()
    prog.update(100)

    return reads_count


def check_read(segments, seq):
    from spoa import poa
    fasta = [seq[int(i.split('-')[0]):int(i.split('-')[1])] for i in segments.split(';')]
    poa(fasta, 1, True, -1, -1, -1, -1, -1)


def scan_raw_chunk(chunk, is_canonical, circ_reads):
    reads_cnt = defaultdict(int)

    ret = []
    short_reads = []

    for read_id, seq in chunk:
        if read_id in circ_reads:
            continue

        # API for short reads
        if len(seq) < 300:
            short_reads.append((read_id, seq))
            continue

        # Remove reads that have ambiguous mapping
        raw_hits = sorted([i for i in env.ALIGNER.map(seq) if i.is_primary], key=lambda x: [x.q_st, x.q_en])
        if len(raw_hits) == 0:
            continue
        elif len(raw_hits) == 1:
            raw_hit = remove_long_insert(raw_hits[0])
            if raw_hit.mlen < len(seq) * .45 or raw_hit.mlen > len(seq) - 50:
                continue
            if raw_hit.q_st < 50 and raw_hit.q_en > len(seq) - 50:
                continue
            circ, junc = find_bsj(seq)
            if junc is None:
                continue

        elif len(raw_hits) == 2:
            head, tail = remove_long_insert(raw_hits[0]), remove_long_insert(raw_hits[1])
            if head.ctg != tail.ctg:
                continue
            if not head.q_st + head.mlen * 0.45 < tail.q_st:
                continue
            if head.r_en - 20 < tail.r_st:
                continue
            if head.q_en < tail.q_st - 50:
                continue
            circ, junc = find_bsj(seq)
            if junc is None or junc < head.q_en - 10 or junc > tail.q_st + 10:
                continue
        else:
            continue

        circ_hits = sorted([remove_long_insert(i) for i in env.ALIGNER.map(circ) if i.is_primary], key=lambda x: [x.q_st, x.q_en])
        if len(circ_hits) == 0:
            continue
        elif len(circ_hits) == 1:
            circ_hit = circ_hits[0]
            if circ_hit.mlen <= max([i.mlen for i in raw_hits]):
                continue
            if min(junc, len(seq) - junc) < 30:
                continue
            if not junc + circ_hit.q_st < len(seq) < junc + circ_hit.q_en:
                continue
            circ_ctg, circ_start, circ_end, circ_strand = circ_hit.ctg, circ_hit.r_st, circ_hit.r_en, circ_hit.strand
            clip_base = circ_hit.q_st + len(seq) - circ_hit.q_en
            cir_exons = get_parital_blocks(circ_hit, len(seq) - junc)
        elif len(circ_hits) == 2:
            head, tail = circ_hits[0], circ_hits[1]
            if head.ctg != tail.ctg or head.strand != tail.strand:
                continue
            if not head.q_st + (head.q_en - head.q_st) * 0.5 < tail.q_st:
                continue
            if head.r_en - 20 < tail.r_st:
                continue
            if head.q_en < tail.q_st - 20:
                continue
            circ_ctg, circ_start, circ_end, circ_strand = head.ctg, tail.r_st, head.r_en, head.strand
            clip_base = abs(tail.q_st - head.q_en)

            head_exons = get_blocks(head)
            tail_exons = get_blocks(tail)

            cir_exons = merge_exons(tail_exons, head_exons)

            circ = circ[tail.q_st:] + circ[:tail.q_st]
        else:
            continue

        if clip_base > 20:
            continue

        # Retrive circRNA positions, convert minimap2 position to real position
        host_strand = find_host_gene(circ_ctg, circ_start, circ_end)
        try:
            ss_site, us_free, ds_free, tmp_signal = find_annotated_signal(circ_ctg, circ_start, circ_end, clip_base, clip_base + 10)
        except Exception as e:
            print(e)
        if ss_site is None:
            ss_site = find_denovo_signal(circ_ctg, circ_start, circ_end, host_strand, tmp_signal,
                                         us_free, ds_free, clip_base, clip_base + 10, 3, True)

        if ss_site is None:
            strand = 'NA'
            ss_id = 'NA'
            correction_shift = 0
        else:
            ss_id, strand, us_shift, ds_shift = ss_site
            circ_start += us_shift
            circ_end += ds_shift
            correction_shift = min(max(us_shift, -us_free), ds_free)

        circ_id = '{}:{}-{}'.format(circ_ctg, circ_start + 1, circ_end)
        cir_exons[0][0] = circ_start
        cir_exons[-1][1] = circ_end

        cir_exon_tag = []
        for cir_exon_start, cir_exon_end, cir_exon_len in cir_exons:
            cir_exon_tag.append('{}-{}|{}'.format(cir_exon_start, cir_exon_end, cir_exon_len))

        circ_seq = circ if circ_strand > 0 else revcomp(circ)
        circ_seq = circ_seq[correction_shift:] + circ_seq[:correction_shift]

        ret.append((
            read_id, circ_id, strand, ','.join(cir_exon_tag), ss_id, '{}|{}-NA'.format(junc, clip_base), 'partial', circ_seq
        ))

        reads_cnt['partial'] += 1

    return reads_cnt, ret, short_reads


def scan_raw_reads(in_file, ref_fasta, gtf_index, intron_index, ss_index, is_canonical, out_dir, prefix, threads):
    import gzip
    import mappy as mp
    from CIRI_long.utils import to_str

    circ_reads = {}
    with open('{}/{}.cand_circ.fa'.format(out_dir, prefix), 'r') as f:
        for line in f:
            read_id = line.rstrip().split()[0].lstrip('>')
            circ_reads[read_id] = 1
            f.readline()

    # Auto detect input format
    is_fastq = 1
    is_gz = 1
    if in_file.endswith('.fa') or in_file.endswith('.fasta'):
        is_fastq = 0
        is_gz = 0
        fq = open(in_file, 'r')
    elif in_file.endswith('.fa.gz') or in_file.endswith('.fasta.gz'):
        is_fastq = 0
        is_gz = 1
        fq = gzip.open(in_file, 'rb')
    elif in_file.endswith('.fq') or in_file.endswith('.fastq'):
        is_gz = 0
        fq = open(in_file, 'r')
    elif in_file.endswith('.fq.gz') or in_file.endswith('.fastq.gz'):
        fq = gzip.open(in_file, 'rb')
    else:
        sys.exit('Wrong format of input')

    # Prepare aligners
    faidx = Faidx(ref_fasta)
    contig_len = faidx.contig_len
    faidx.close()

    aligner = mp.Aligner(ref_fasta, n_threads=threads, preset='splice')

    jobs = []
    pool = Pool(threads, env.initializer, (aligner, contig_len, aligner, gtf_index, intron_index, ss_index))

    # Init jobs
    chunk = []
    chunk_size = 1000
    for line in fq:
        if is_gz:
            header = to_str(line).rstrip().split(' ')[0]
            seq = to_str(fq.readline()).rstrip()
        else:
            header = line.rstrip().split(' ')[0]
            seq = fq.readline().rstrip()

        if is_fastq:
            header = header.lstrip('@')
            fq.readline()
            fq.readline()
        else:
            header = header.lstrip('>')

        # Split into chunks
        chunk.append((header, seq))
        if len(chunk) == chunk_size:
            jobs.append(pool.apply_async(scan_raw_chunk, (chunk, is_canonical, circ_reads)))
            chunk = []

    if len(chunk) > 0:
        jobs.append(pool.apply_async(scan_raw_chunk, (chunk, is_canonical, circ_reads)))
    pool.close()

    # Receive results
    reads_cnt = defaultdict(int)

    prog = ProgressBar()
    prog.update(0)
    finished_job = 0

    short_reads = []
    with open('{}/{}.low_confidence.fa'.format(out_dir, prefix), 'w') as out:
        for job in jobs:
            tmp_cnt, tmp_ret, tmp_short = job.get()
            for key, value in tmp_cnt.items():
                reads_cnt[key] += value
            short_reads += tmp_short

            for read_id, circ_id, strand, cir_exon_tag, ss_id, clip_info, segments, circ_seq in tmp_ret:
                out.write('>{}\t{}\t{}\t{}\t{}\t{}\t{}\n{}\n'.format(
                    read_id, circ_id, strand, cir_exon_tag, ss_id, clip_info, segments, circ_seq
                ))
            finished_job += 1

            prog.update(100 * finished_job // len(jobs))

    prog.update(100)
    pool.join()

    return reads_cnt, short_reads


# def recover_raw_chunk(chunk, is_canonical):
#     reads_cnt = defaultdict(int)
#
#     ret = []
#     for read_id, seq in chunk:
#         raw_hits = ALIGNER.map(seq)
#         if raw_hits is None:
#             continue
#         raw_hits = sorted([i for i in raw_hits if i.is_primary], key=lambda x: [x.q_st, x.q_en])
#         if len(raw_hits) == 0:
#             continue
#
#         if len(raw_hits) == 1:
#             raw_hit = raw_hits[0]
#             if raw_hit.q_en - raw_hit.q_st < len(seq) * .45:
#                 continue
#             if raw_hit.q_en - raw_hit.q_st > len(seq) - 50:
#                 continue
#             if raw_hit.q_st < 50 and len(seq) - raw_hit.q_en < 50:
#                 continue
#
#             circ, junc = find_bsj(seq)
#             if junc is None:
#                 continue
#             circ_hits = ALIGNER.map(circ)
#             if circ_hits is None:
#                 continue
#             circ_hits = sorted([i for i in circ_hits if i.is_primary], key=lambda x: [20, x.q_en])
#             if len(circ_hits) == 0:
#                 continue
#
#             if len(circ_hits) == 1:
#                 circ_hit = circ_hits[0]
#                 if circ_hit.mlen <= raw_hit.mlen:
#                     continue
#                 if min(junc, len(seq) - junc) < 30:
#                     continue
#                 if not junc + circ_hit.q_st < len(seq) < junc + circ_hit.q_en:
#                     continue
#                 circ_ctg, circ_start, circ_end = circ_hit.ctg, circ_hit.r_st, circ_hit.r_en
#                 clip_base = circ_hit.q_st + len(seq) - circ_hit.q_en
#
#             elif len(circ_hits) == 2:
#                 head, tail = circ_hits[0], circ_hits[1]
#                 if head.ctg != tail.ctg:
#                     continue
#                 if not head.q_st + (head.q_en - head.q_st) * 0.5 < tail.q_st:
#                     continue
#                 if head.r_en - 20 < tail.r_st:
#                     continue
#                 if head.q_en < tail.q_st - 50:
#                     continue
#                 circ_ctg, circ_start, circ_end = head.ctg, tail.r_st, head.r_en
#                 clip_base = abs(tail.q_st - head.q_en)
#             else:
#                 continue
#
#         elif len(raw_hits) == 2:
#             head, tail = raw_hits[0], raw_hits[1]
#             if head.ctg != tail.ctg:
#                 continue
#             if not head.q_st + (head.q_en - head.q_st) * 0.5 < tail.q_st:
#                 continue
#             if head.r_en - 20 < tail.r_st:
#                 continue
#             if head.q_en < tail.q_st - 50:
#                 continue
#             circ_ctg, circ_start, circ_end = head.ctg, tail.r_st, head.r_en
#             clip_base = abs(tail.q_st - head.q_en)
#
#         else:
#             continue
#
#         # Retrive circRNA positions, convert minimap2 position to real position
#         ss_site, us_free, ds_free = search_splice_signal(circ_ctg, circ_start, circ_end, clip_base)
#         if ss_site is None:
#             continue
#
#         ss_id, strand, us_shift, ds_shift = ss_site
#         circ_start += us_shift
#         circ_end += ds_shift
#
#         # if is_canonical: keep canonical splice site only
#         ss = ss_id.split('|')[0]
#         if ss_site is None or (is_canonical and ss[-1] == '*'):
#             continue
#
#         circ_id = '{}:{}-{}'.format(circ_ctg, circ_start + 1, circ_end)
#
#     return reads_cnt, ret
#
#
# def recover_raw_reads(short_reads, ref_fasta, ss_index, is_canonical, out_dir, prefix, threads):
#     from bwapy import BwaAligner
#
#     faidx = Faidx(ref_fasta)
#     contig_len = faidx.contig_len
#
#     options = '-x ont2d -T 19'
#     bwa_aligner = Aligner(BwaAligner(ref_fasta, options=options))
#
#     chunk_size = 250
#     jobs = []
#     pool = Pool(threads, initializer, (bwa_aligner, faidx, ss_index, contig_len))
#     for reads in grouper(short_reads, chunk_size):
#         chunk = [i for i in reads if i is not None]
#         jobs.append(pool.apply_async(recover_raw_chunk, (chunk, is_canonical)))
#     pool.close()
#
#     prog = ProgressBar()
#     prog.update(0)
#     finished_job = 0
#
#     reads_cnt = defaultdict(int)
#     with open('{}/{}.low_confidence.fa'.format(out_dir, prefix), 'a') as out:
#         for job in jobs:
#             finished_job += 1
#
#             tmp_cnt, tmp_ret = job.get()
#             for key, value in tmp_cnt.items():
#                 reads_cnt[key] += value
#
#             prog.update(100 * finished_job // len(jobs))
#
#     pool.join()
#     prog.update(100)
#
#     faidx.close()
#
#     return reads_cnt



#!/home/zhangjy/.virtualenvs/Benchmarking/bin/python
import os
import sys
import pickle
import json
from collections import defaultdict


def call(args):
    from CIRI_long.logger import get_logger
    from CIRI_long.utils import check_file, check_dir
    from CIRI_long.align import index_annotation, index_circ
    from CIRI_long.find_ccs import find_ccs_reads, load_ccs_reads
    from CIRI_long.find_bsj import scan_ccs_reads, recover_ccs_reads
    from CIRI_long.find_bsj import scan_raw_reads
    from CIRI_long.pipeline import run_ccs

    lib_path = os.path.dirname(os.path.split(os.path.realpath(__file__))[0]) + '/libs'
    os.environ['PATH'] = lib_path + ':' + os.environ['PATH']
    os.chmod(lib_path + '/ccs', 0o755)

    if args.input is None or args.output is None:
        sys.exit('Please provide input and output file, run CIRI-long using -h or --help for detailed information.')
    if args.reference is None:
        sys.exit('Please specific FASTA of reference genome')

    # Check parameters
    in_file = check_file(args.input)
    gtf_file = None if args.gtf is None else check_file(args.gtf)
    circ_file = None if args.circ is None else check_file(args.circ)
    out_dir = check_dir(args.output)
    ref_fasta = check_file(args.reference)
    check_dir(out_dir + '/tmp')
    prefix = args.prefix
    threads = int(args.threads)
    debugging = args.debug
    is_canonical = args.canonical

    logger = get_logger('CIRI-long', fname='{}/{}.log'.format(out_dir, prefix), verbosity=debugging)
    logger.info('----------------- Input paramters ------------------')
    logger.info('Input reads: ' + os.path.basename(in_file))
    logger.info('Output directory: ' + os.path.basename(out_dir))
    logger.info('Multi threads: {}'.format(args.threads))
    logger.info('----------------- Calling circRNAs -----------------')

    # Scan for repeats and CCS
    reads_count = defaultdict(int)
    is_fast = 0

    if not debugging and os.path.exists('{}/tmp/{}.ccs.fa'.format(out_dir, prefix)) and os.path.exists('{}/tmp/{}.raw.fa'.format(out_dir, prefix)):
        logger.info('Step 1 - Loading circRNA candidates in previous run')
        ccs_seq = load_ccs_reads(out_dir, prefix)
        reads_count['consensus'] = len(ccs_seq)
    else:
        try:
            run_ccs(in_file, out_dir, prefix, threads, debugging)
            ccs_seq = load_ccs_reads(out_dir, prefix)
            reads_count['consensus'] = len(ccs_seq)
            is_fast = 1
        except Exception as e:
            logger.warn('Failed to run ccs command in fast mode, falling back to slower python version.')
            total_reads, ro_reads, ccs_seq = find_ccs_reads(in_file, out_dir, prefix, threads, debugging)
            reads_count['total'] = total_reads
            reads_count['consensus'] = ro_reads
            is_fast = 0

    if is_fast == 1 and reads_count['consensus'] == 0:
        logger.warn('Failed to run ccs command in fast mode, try again in slower python version.')
        total_reads, ro_reads, ccs_seq = find_ccs_reads(in_file, out_dir, prefix, threads, debugging)
        reads_count['total'] = total_reads
        reads_count['consensus'] = ro_reads

    if 'total' in reads_count:
        logger.info('Total Reads: {}'.format(reads_count['total']))
    logger.info('Cyclic Consensus Reads: {}'.format(reads_count['consensus']))

    # generate index of splice site and annotation
    if gtf_file is None and circ_file is None:
        logger.warn('No annotation provided, entering \'De novo\' mode')
        gtf_idx, ss_idx = None, None
    else:
        idx_file = out_dir + '/tmp/ss.idx'
        if os.path.exists(idx_file):
            logger.info('Loading pre-built splice site index from: {}'.format(idx_file))
            with open(idx_file, 'rb') as idx:
                gtf_idx, intron_idx, ss_idx = pickle.load(idx)
        else:
            if gtf_file is not None:
                gtf_idx, intron_idx, ss_idx = index_annotation(gtf_file)
            else:
                gtf_idx, intron_idx, ss_idx = None, None, None
            if circ_file is not None:
                ss_idx = index_circ(circ_file, ss_idx)

            with open(idx_file, 'wb') as idx:
                pickle.dump([gtf_idx, intron_idx, ss_idx], idx, -1)

    # Find circRNAs
    logger.info('Step 2.1 - Find circRNAs from CCS reads')
    tmp_cnt, short_seq = scan_ccs_reads(ccs_seq, ref_fasta, ss_idx, gtf_idx, intron_idx, is_canonical, out_dir, prefix, threads)
    for key, value in tmp_cnt.items():
        reads_count[key] += value

    # Recover short reads
    logger.info('Step 2.2 - Recover short CCS reads')
    tmp_cnt = recover_ccs_reads(short_seq, ref_fasta, ss_idx, gtf_idx, intron_idx, is_canonical, out_dir, prefix, threads)
    for key, value in tmp_cnt.items():
        reads_count[key] += value

    # Find BSJs
    logger.info('Step 3 - Find circRNAs with partial structure')
    tmp_cnt, short_seq = scan_raw_reads(in_file, ref_fasta, gtf_idx, intron_idx, ss_idx, is_canonical, out_dir, prefix, threads)
    for key, value in tmp_cnt.items():
        reads_count[key] += value

    logger.info('Raw unmapped: {}'.format(reads_count['raw_unmapped']))
    logger.info('CCS mapped: {}'.format(reads_count['ccs_mapped']))
    logger.info('BSJ: {}'.format(reads_count['bsj']))
    logger.info('Splice signal: {}'.format(reads_count['signal']))
    logger.info('Partial reads: {}'.format(reads_count['partial']))

    with open('{}/{}.json'.format(out_dir, prefix), 'w') as f:
        json.dump(reads_count, f)

    logger.info('Calling circRNAs finished!')


def collapse(args):
    from CIRI_long.logger import get_logger
    from CIRI_long.utils import check_file, check_dir
    from CIRI_long.align import index_annotation, index_circ
    from CIRI_long import collapse

    if args.input is None or args.output is None:
        sys.exit('Please provide input and output file, run CIRI-long using -h or --help for detailed information.')

    in_file = check_file(args.input)
    out_dir = check_dir(args.output)
    check_dir(out_dir + '/tmp')
    prefix = args.prefix

    gtf_file = None if args.gtf is None else check_file(args.gtf)
    circ_file = None if args.circ is None else check_file(args.circ)
    ref_fasta = check_file(args.reference)

    threads = int(args.threads)
    debugging = args.debug
    is_canonical = args.canonical

    logger = get_logger('CIRI-long', fname='{}/{}.log'.format(out_dir, prefix), verbosity=debugging)
    logger.info('----------------- Input paramters ------------------')
    logger.info('Input reads: ' + os.path.basename(in_file))
    logger.info('Output directory: ' + os.path.basename(out_dir))
    logger.info('Multi threads: {}'.format(args.threads))
    logger.info('-------------- Collapse circular reads -------------')

    # generate index of splice site and annotation
    if gtf_file is None and circ_file is None:
        logger.warn('No annotation provided, entering \'De novo\' mode')
        gtf_idx, ss_idx = None, None
    else:
        idx_file = out_dir + '/tmp/ss.idx'
        if os.path.exists(idx_file):
            logger.info('Loading pre-built splice site index from: {}'.format(idx_file))
            with open(idx_file, 'rb') as idx:
                gtf_idx, intron_idx, ss_idx = pickle.load(idx)
        else:
            if gtf_file is not None:
                gtf_idx, intron_idx, ss_idx = index_annotation(gtf_file)
            else:
                gtf_idx, intron_idx, ss_idx = None, None, None
            if circ_file is not None:
                ss_idx = index_circ(circ_file, ss_idx)

            with open(idx_file, 'wb') as idx:
                pickle.dump([gtf_idx, intron_idx, ss_idx], idx, -1)

    # Load reads
    cand_reads = collapse.load_cand_circ(in_file)

    # Consensus reads
    corrected_file = '{}/tmp/{}.corrected.pkl'.format(out_dir, prefix)
    if not debugging and os.path.exists(corrected_file):
        logger.info('Step 1 - Loading clustered circular reads in previous run')
        with open(corrected_file, 'rb') as pkl:
            circ_num, corrected_reads = pickle.load(pkl)
    else:
        logger.info('Step 1 - Clustering candidate circular reads')
        # Cluster reads
        reads_cluster = collapse.cluster_reads(cand_reads)
        logger.info('Circular reads clusters: {}'.format(len(reads_cluster)))

        # Generate consensus reads
        circ_num, corrected_reads = collapse.correct_reads(reads_cluster, ref_fasta, gtf_idx, intron_idx, ss_idx, threads)
        with open(corrected_file, 'wb') as pkl:
            pickle.dump([circ_num, corrected_reads], pkl, -1)
        logger.info('Corrected clusters: {}, {}/{}/{}/{} annotated/denovo/lariat/unknown'.format(
            len(corrected_reads), circ_num['Annotated'], circ_num['Denovo signal'],
            circ_num['High confidence lariat'], circ_num['Unknown signal']))

    logger.info('Step 2 - Calculating expression matrix')
    circ_cnt = collapse.cal_exp_mtx(cand_reads, corrected_reads, ref_fasta, gtf_idx, out_dir, prefix)
    logger.info('Final circRNAs: {}'.format(circ_cnt))

    # # Find circRNAs again
    # logger.info('Correct circRNAs from consensus reads!')
    # corrected_circ, short_reads = scan_corrected_reads(corrected_reads, ref_fasta, ss_idx, threads)
    #
    # # Recover short circRNAs
    # logger.info('Recover short circRNAs from consensus reads!')
    # corrected_circ += recover_corrected_reads(short_reads, ref_fasta, ss_idx)
    #
    # logger.info('Clustered circRNAs: {}'.format(len(corrected_reads)))
    # logger.info('Corrected circRNAs: {}'.format(len(corrected_circ)))
    #
    # with open('{}/{}_circ.pkl'.format(out_dir, prefix), 'wb') as idx:
    #     pickle.dump(corrected_circ, idx, -1)

    logger.info('Correction of Back-Spliced Junctions finished!')


def main():
    import argparse
    from CIRI_long.version import __version__

    # Init parser
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s v{version}'.format(version=__version__))

    # Init subparsers
    subparsers = parser.add_subparsers(help='commands')

    # Parsers for circRNA calling
    call_parser = subparsers.add_parser('call')
    call_parser.add_argument('-i', '--in', dest='input', metavar='READS', default=None,
                             help='Input reads.fq.gz', )
    call_parser.add_argument('-o', '--out', dest='output', metavar='DIR', default=None,
                             help='Output directory, default: ./', )
    call_parser.add_argument('-r', '--ref', dest='reference', metavar='REF', default=None,
                             help='Reference genome FASTA file', )
    call_parser.add_argument('-p', '--prefix', dest='prefix', metavar='PREFIX', default="CIRI-long",
                             help='Output sample prefix, (default: %(default)s)', )
    call_parser.add_argument('-a', '--anno', dest='gtf', metavar='GTF', default=None,
                             help='Genome reference gtf, (optional)', )
    call_parser.add_argument('-c', '--circ', dest='circ', metavar='CIRC', default=None,
                             help='Additional circRNA annotation in bed/gtf format, (optional)', )
    call_parser.add_argument('--canonical', dest='canonical', default=True, action='store_true',
                             help='Use canonical splice signal (GT/AG) only, default: %(default)s)')
    call_parser.add_argument('-t', '--threads', dest='threads', metavar='INT', default=os.cpu_count(),
                             help='Number of threads, (default: use all cores)', )
    call_parser.add_argument('--debug', dest='debug', default=False, action='store_true',
                             help='Run in debugging mode, (default: %(default)s)', )
    call_parser.set_defaults(func=call)

    # Parsers for concat circular reads
    collapse_parser = subparsers.add_parser('collapse')
    collapse_parser.add_argument('-i', '--in', dest='input', metavar='LIST', default=None,
                                 help='Input list of CIRI-long results', )
    collapse_parser.add_argument('-o', '--out', dest='output', metavar='DIR', default=None,
                                 help='Output directory, default: ./', )
    collapse_parser.add_argument('-p', '--prefix', dest='prefix', metavar='PREFIX', default="CIRI-long",
                                 help='Output sample prefix, (default: %(default)s)', )
    collapse_parser.add_argument('-r', '--ref', dest='reference', metavar='REF', default=None,
                                 help='Reference genome FASTA file', )
    collapse_parser.add_argument('-a', '--anno', dest='gtf', metavar='GTF', default=None,
                                 help='Genome reference gtf, (optional)', )
    collapse_parser.add_argument('-c', '--circ', dest='circ', metavar='CIRC', default=None,
                                 help='Additional circRNA annotation in bed/gtf format, (optional)', )
    collapse_parser.add_argument('--canonical', dest='canonical', default=True, action='store_true',
                                 help='Use canonical splice signal (GT/AG) only, default: %(default)s)')
    collapse_parser.add_argument('-t', '--threads', dest='threads', metavar='INT', default=os.cpu_count(),
                                 help='Number of threads, (default: use all cores)', )
    collapse_parser.add_argument('--debug', dest='debug', default=False, action='store_true',
                                 help='Run in debugging mode, (default: %(default)s)', )

    collapse_parser.set_defaults(func=collapse)

    # Parse parsers
    args = parser.parse_args()

    # Run function
    try:
        args.func(args)
    except AttributeError as e:
        parser.print_help()


if __name__ == '__main__':
    main()
